import logging
import os
from collections import defaultdict
from functools import reduce
from pathlib import Path

from rcounting.reddit_interface import subreddit

printer = logging.getLogger(__name__)
default_filename = os.path.dirname(__file__) / Path("aliases.txt")


def read_aliases(filename=default_filename):
    with open(filename, "r", encoding="utf8") as f:
        alias_list = f.readlines()
    alias_list = [line.replace(" ", "").strip().split(",") for line in alias_list]
    alias_dict = {}
    for aliases in alias_list:
        for alias in aliases:
            alias_dict[alias] = aliases[0]
    return alias_dict


def normalise_aliases(alias_dict):
    aliases = {}
    for alias in alias_dict:
        aliases[alias.lower()] = alias_dict[alias]
    return aliases


def to_csv(mydict):
    s = ""
    for key in sorted(mydict.keys(), key=lambda x: x.lower()):
        vals = mydict[key]
        s += ",".join([key] + sorted(set(vals) - set([key]), key=lambda x: x.lower())) + "\n"
    return s


def read_aliases_from_wiki(page):
    wiki_page = subreddit.wiki[page]
    document = wiki_page.content_md.replace("\r\n", "\n").replace("*", "")
    lines = document.split("\n")
    aliases = [y for line in lines if len(y := line.split(":")) == 2]
    new_aliases = {}
    for canonical_name, current_aliases in aliases:
        alias_list = current_aliases.replace(" ", "").split(",")
        new_aliases[canonical_name] = alias_list
    return new_aliases


def sync_aliases(alias_dict, wiki_aliases):
    """The meat of the logic for merging two alias dicts.

    The first is assumed to be in the form key -> canonical name, while the
    second is in the form canonical_name -> list of aliases.

    Returns a dict of the form canonical_name -> list of aliases, with the following merge logic:

    - If a collection of aliases is present in both dictionaries,
      merge the two and use the merged collection.
    - If a collection of aliases is only present in one of the two dictionaries,
      add that directly to the result.
    - If the canonical name is different between the stored aliases and the wiki aliases,
      use the wiki name.
    """
    reverse_dict = defaultdict(list)
    for key in alias_dict:
        reverse_dict[alias_dict[key]].append(key)
    updated_aliases = {}
    overlapping_aliases = set()
    for key in wiki_aliases:
        target_keys = [key.lower()] + [x.lower() for x in wiki_aliases[key]]
        target_keys = {alias_dict[x] for x in target_keys if x in alias_dict}
        overlapping_aliases |= target_keys
        old_aliases = reduce(
            lambda x, y: x | set(y), [reverse_dict[x] for x in target_keys], set()
        )
        new_aliases = list(set(wiki_aliases[key]))
        lowercase = [x.lower() for x in new_aliases]
        for old_alias in old_aliases:
            if old_alias.lower() not in lowercase:
                new_aliases.append(old_alias)
                lowercase.append(old_alias.lower())
        updated_aliases[key] = new_aliases
    updated_aliases = updated_aliases | {
        k: v for k, v in reverse_dict.items() if k not in overlapping_aliases
    }
    return updated_aliases


def write_aliases(aliases, filename=default_filename):
    csv_text = to_csv(aliases)
    with open(filename, "w", encoding="utf8") as f:
        f.write(csv_text)


def update_aliases(filename=default_filename):
    """The main interface to the alias updates. Reads the stored alias file,
    loads the one from the wiki page, merges the two and overwrites the old
    alias file. The internal alias dict is then updated, so that `apply_alias`
    will work as expected right after calling this function.
    """

    alias_dict = read_aliases(filename=filename)
    wiki_aliases = read_aliases_from_wiki("aliases")
    try:
        updated_aliases = sync_aliases(alias_dict, wiki_aliases)
    # Pylint complains about a broad except here which is fair enough; the idea
    # is that at any signs of error we abort the operation and just keep going
    # with the old aliases dict. That seems safer than trying to recover and
    # ending up in an unknown state.
    except Exception as e:  # pylint: disable=broad-except
        printer.error("Something went wrong while fetching aliases from wiki page. Aborting")
        printer.error(e, exc_info=True)
        return
    write_aliases(updated_aliases, filename=filename)
    # Pylint complains about this global. For now, it seems to be the easiest
    # way to share state between this function and `apply alias`.
    global _alias_dict  # pylint: disable=global-statement
    _alias_dict = normalise_aliases(read_aliases())


_alias_dict = normalise_aliases(read_aliases())


def apply_alias(username):
    return _alias_dict[username.lower()] if username.lower() in _alias_dict else username


mods = [
    "Z3F",
    "949paintball",
    "zhige",
    "atomicimploder",
    "ekjp",
    "TheNitromeFan",
    "davidjl123",
    "rschaosid",
    "KingCaspianX",
    "Urbul",
    "Zaajdaeon",
]


def is_mod(username):
    return username in mods


ignored_counters = [
    "LuckyNumber-Bot",
    "CountingStatsBot",
    "CountingHelper",
    "WikiSummarizerBot",
    "InactiveUserDetector",
    "alphabet_order_bot",
    "exclaim_bot",
]
ignored_counters = [x.lower() for x in ignored_counters]
banned_counters = ["[deleted]", "Franciscouzo", "None"]


def is_ignored_counter(username):
    return username.lower() in ignored_counters


def is_banned_counter(username):
    return username in banned_counters



from datetime import timedelta

import matplotlib.colors as mcolors
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cycler import cycler
from pandas.plotting import register_matplotlib_converters

from rcounting.analysis import fft_kde
from rcounting.counters import apply_alias, is_banned_counter
from rcounting.units import DAY, HOUR, MINUTE

register_matplotlib_converters()


def format_x_date_month(ax):
    months = mdates.MonthLocator()  # every month
    quarters = mdates.MonthLocator([1, 4, 7, 10])
    monthFmt = mdates.DateFormatter("%b %Y")
    ax.xaxis.set_major_locator(quarters)
    ax.xaxis.set_major_formatter(monthFmt)
    ax.xaxis.set_minor_locator(months)


def parts_vs_counts(df):
    k_parts = df.groupby("username")["submission_id"].nunique()
    hoc = df.groupby("username")["submission_id"].count()
    combined = pd.merge(k_parts, hoc, left_index=True, right_index=True)
    combined.columns = ["k_parts", "total_counts"]
    combined = combined.query("k_parts > 10")
    assert combined is not None, "Error, querying too small a dataframe"
    linear_model = np.polyfit(np.log10(combined["k_parts"]), np.log10(combined["total_counts"]), 1)
    print(linear_model)
    axis = np.linspace(1, combined["k_parts"].max(), endpoint=True)

    plt.scatter(combined["k_parts"], combined["total_counts"], alpha=0.7)
    plt.plot(
        axis, 10 ** (np.poly1d(linear_model)(np.log10(axis))), linestyle="--", color="0.3", lw=2
    )
    plt.xlabel("Threads participated in ")
    plt.ylabel("Total counts made")
    plt.yscale("log")
    plt.xscale("log")
    plt.xlim(left=10)
    plt.ylim(bottom=10)
    plt.savefig("parts_vs_counts.png", dpi=300, bbox_inches="tight")


def hoc_by_time(df, n=25, ax=None):
    if ax is None:
        _, ax = plt.subplots(1, figsize=(10, 6))

    if df.index.inferred_type != "datetime64" and "timestamp" in df.columns:
        df["date"] = pd.to_datetime(df["timestamp"], unit="s")
        df.set_index("date", inplace=True)
        df.sort_index(inplace=True)

    df["username"] = df["username"].apply(apply_alias)
    top_counters = [
        x
        for x in df.groupby("username").size().sort_values(ascending=False).index
        if not is_banned_counter(x)
    ][:n]
    filtered = df.query("username in @top_counters")
    cumulative = pd.get_dummies(filtered["username"]).resample("12h").sum().expanding().sum()
    cumulative[top_counters].plot(ax=ax)
    plt.legend(bbox_to_anchor=(1.05, 1.07))
    ax.set_ylabel("Cumulative counts")
    ax.set_xlabel("")
    ax.set_ylim(bottom=0)
    return ax


def speedrun_histogram(df, n=3):
    bins = np.arange(0, 21)
    df = df.copy()
    df["dt"] = df["timestamp"].diff()
    counters = df.query("dt < 20").groupby("username")["dt"].mean().sort_values().index
    fig, axes = plt.subplots(n, sharex=True, sharey=True)
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    for idx, counter in enumerate(counters[:n]):
        ax = axes[idx]
        ax.hist(
            df.query("username == @counter")["dt"],
            bins=bins,
            alpha=0.6,
            label=counter,
            density=True,
            color=colors[idx],
            edgecolor="k",
        )
        ax.legend()
    ax.set_xlim(0, 11)
    ax.set_xticks([0.5, 5.5, 10.5])
    ax.set_xticklabels(["0", "5", "10"])
    return fig


def time_of_day_histogram(df, ax, n=4):
    bins = np.linspace(0, DAY, 24 * 12 + 1)
    df = df.copy()
    df["time_of_day"] = df["timestamp"].astype(int) % DAY
    top_counters = df["username"].value_counts().index[:n]
    ax.hist(df["time_of_day"], bins=bins, alpha=0.8, label="total", color="C3", edgecolor="k")
    for counter in top_counters:
        data = df.query("username==@counter")["time_of_day"]
        ax.hist(data, bins=bins, alpha=0.7, label=counter, edgecolor="k")
    ax.set_xlim(0, DAY + 1)
    intervals = range(0, 25, 3)
    ax.set_xticks([x * HOUR for x in intervals])
    ax.set_xticklabels(["f{x:02d}:00" for x in intervals])
    ax.legend()
    ax.set_ylabel("Number of counts per 5 min interval")
    return ax


def time_of_day_kde(
    df, ax, n_counters=4, time_resolution=MINUTE / 2, show_total=True, normalize=True
):
    alpha = 0.8
    sigma = 0.02
    nbins = int(DAY / time_resolution)
    df = df[["username", "timestamp"]].copy()
    df["time_of_day"] = df["timestamp"].astype(int) % DAY
    counts = df["username"].value_counts()
    top_counters = counts.index[:n_counters]
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    if show_total:
        x, kde = fft_kde(df["time_of_day"], nbins, kernel="normal_distribution", sigma=sigma)
        if normalize:
            kde *= len(df)
        ax.fill_between(x, kde, label="All Counts", color="0.8")
    for idx, counter in enumerate(top_counters):
        data = df.query("username==@counter")["time_of_day"]
        x, kde = fft_kde(data, nbins, kernel="normal_distribution", sigma=sigma)
        if normalize:
            kde *= counts.loc[counter]
        ax.fill_between(x, kde, color=colors[idx], alpha=alpha)
        ax.plot(x, kde, label=counter, color=colors[idx], lw=2)
    intervals = range(0, 25, 3)
    ax.set_xlim(0, DAY + 1)
    ax.set_xticks([x * HOUR for x in intervals])
    ax.set_xticklabels([f"{x:02d}:00" for x in intervals])
    ax.set_ylim(bottom=0)
    ax.legend()
    ax.set_xlabel("Time of day [UTC]")
    ax.set_ylabel("Counts per second")
    return ax


def plot_get_time(df, ax, **kwargs):
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][:2]
    cc = cycler(colors) * cycler(marker=list("vs"))
    ax.set_prop_cycle(cc)
    modstring = {False: "Non-mod", True: "Mod"}
    get_type = {"get": "Get", "assist": "Assist"}

    for count_type in ["get", "assist"]:
        for modness in [False, True]:
            subset = df.query("is_moderator == @modness and count_type == @count_type")
            ax.plot(
                subset["timestamp"],
                subset["elapsed_time"],
                linestyle="None",
                label=f"{get_type[count_type]} by {modstring[modness]}",
                alpha=0.8,
                **kwargs,
            )

    one_day = timedelta(1)
    start_date, end_date = df["timestamp"].min(), df["timestamp"].max()
    ax.set_xlim(left=start_date - one_day, right=end_date + one_day)
    format_x_date_month(ax)
    ax.set_ylim(0, 30)
    ax.set_ylabel("Elapsed time for assists and gets [s]")
    ax.legend(loc="upper right")
    return ax


def simulate_alpha(color, alpha):
    """Find the average of the current color and pure white, weighted by alpha"""
    white = np.array((1, 1, 1))
    return tuple(np.array(mcolors.to_rgb(color)) * alpha + (1 - alpha) * white)


def make_time_axis(ax):
    """Take an x axis that runs from 0 to 86400 seconds and make ticks at
    sensible intervals with sensible names

    """
    ticks, labels = zip(*[(x * HOUR, f"{x:02d}:00") for x in range(0, 25, 3)])
    ax.set_xlim(0, DAY + 1)
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels)
    ax.set_ylim(bottom=0)



"""
Provide an authorised interface to the reddit api.
Look for authorisation first in the environment variables, then in a config file.
If none can be found, prompt the user to authorise."""
import configparser
import os
import random
import socket
from importlib.metadata import version

import praw
import praw.exceptions

# The tools work with OAuth access and refresh tokens, so you need to grant
# access. Once you've done that, they token will be stored in a file for future
# use.
_CLIENT_ID = "S2MLNKwl9tpgaS5jUtlIIQ"
_USER_AGENT = f"rcounting_tools/v{version('rcounting')} by u/CutOnBumInBandHere9"


def get_refresh_token():
    """
    Prompt the user to authorise the program with reddit,
    and open a tcp connection to grab the returned refresh token. Return this token to the caller.

    Ask for permission to read posts & wiki pages, and edit wiki pages.
    """
    scopes = ["read", "wikiread", "wikiedit", "modposts", "submit"]

    auth_reddit = praw.Reddit(
        client_id=_CLIENT_ID,
        user_agent=_USER_AGENT,
        redirect_uri="http://localhost:8080",
        client_secret=None,
    )
    print(
        "This is the first time you're using the rcounting tools.",
        "You need to allow the program to interact with reddit on your behalf.",
        "For logging, it needs to be able to read wiki pages and posts, "
        "to update the thread directory it needs permission to edit wiki pages, "
        "and to create and pin an FTF post it needs permission to post and mod submissions",
        sep="\n",
    )
    state = str(random.randint(0, 65000))
    url = auth_reddit.auth.url(scopes=scopes, state=state)
    print(f"Please open this url in a web browser, and follow the instructions there: {url}")

    client = receive_connection()
    data = client.recv(1024).decode("utf-8")
    param_tokens = data.split(" ", 2)[1].split("?", 1)[1].split("&")
    params = dict([token.split("=") for token in param_tokens])

    if state != params["state"]:
        send_message(
            client,
            f"State mismatch. Expected: {state} Received: {params['state']}",
        )
        return 1
    if "error" in params:
        send_message(client, params["error"])
        return 1

    token = auth_reddit.auth.authorize(params["code"])
    send_message(client, "Successfully retrieved refresh token! You can close this window now")
    return token


def receive_connection():
    """
    Wait for and then return a connected socket..

    Opens a TCP connection on port 8080, and waits for a single client.

    """
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(("localhost", 8080))
    server.listen(1)
    client = server.accept()[0]
    server.close()
    return client


def send_message(client, message):
    """Send message to client and close the connection."""
    print(message)
    client.send(f"HTTP/1.1 200 OK\r\n\r\n{message}".encode("utf-8"))
    client.close()


refresh_token = os.getenv("praw_refresh_token")
if refresh_token is None:
    module_dir = os.path.dirname(__file__)
    credentials_file = os.path.join(module_dir, "credentials.ini")
    if not os.path.isfile(credentials_file):
        refresh_token = get_refresh_token()
        with open(credentials_file, "w", encoding="utf8") as f:
            print(f"[tokens]\nrefresh_token = {refresh_token}", file=f)
    config = configparser.ConfigParser()
    config.read(credentials_file)
    refresh_token = config["tokens"]["refresh_token"]

reddit = praw.Reddit(
    client_id=_CLIENT_ID, user_agent=_USER_AGENT, client_secret=None, refresh_token=refresh_token
)

subreddit = reddit.subreddit("counting")
reddit.validate_on_submit = True


def extract_from_short_link(url):
    try:
        comment = reddit.comment(url=url)
        return comment.submission.id, comment.id
    except praw.exceptions.InvalidURL:
        submission = reddit.submission(url=url)
        return submission.id, None



import functools
import logging
from typing import Callable

import numpy as np
import pandas as pd

from rcounting import counters, parsing
from rcounting import thread_navigation as tn
from rcounting import utils
from rcounting.models import comment_to_dict

from .rules import default_rule
from .validate_count import base_n_count
from .validate_form import permissive

printer = logging.getLogger(__name__)


def make_title_updater(comment_to_count):
    @functools.wraps(comment_to_count)
    def wrapper(_, chain):
        title = chain[-1].title
        return comment_to_count(parsing.body_from_title(title))

    return wrapper


class SideThread:
    """A side thread class, which consists of a validation part and an update
    part In addition to checking whether a collection of counts is valid
    according to the side thread rule, the class can take a mapping
    comment->count and using this try and identify when errors were made in the
    chain. The class will also attempt to determine how many total counts have
    been made in a given side thread using one of:

    - The comment->count mapping to determine the current count, which is then
    applied to the submission title

    - The update_function parameter, which takes in the current state and
    returns the total number of counts. Sensible approaches for doing this are
    either parsing the current state from the title if it's different from the
    comment->count mapping, or traversing the chain of comments until the last
    known state is reached, and adding on all the comments encountered along
    the way. This is useful for threads which don't have a constant number of
    counts between gets, e.g. tug of war.

    - A standard thread length

    The approaches are listed in low->high priority, so if more than one
    approach is supplied the highest priority one is used.

    """

    def __init__(
        self,
        rule=default_rule,
        form: Callable[[str], bool] = permissive,
        length: int | None = None,
        comment_to_count: Callable[[str], int] | None = None,
        update_function=None,
    ):
        self.form = form
        self.rule = rule
        self.history = None
        self.comment_to_count = None
        if comment_to_count is not None:
            self.comment_to_count = comment_to_count
            self.update_count = make_title_updater(comment_to_count)
        if update_function is not None:
            self.update_count = update_function
        if length is not None or (comment_to_count is None and update_function is None):
            self.length = length if length is not None else 1000
            self.update_count = self.update_from_length

    def update_from_length(self, old_count, chain):
        if self.length is not None:
            return old_count + self.length * (len(chain))
        return None

    def is_valid_thread(self, history):
        mask = self.rule.is_valid(history)
        if mask.all():
            return (True, "")
        return (False, history.loc[~mask, "comment_id"].iloc[0])

    def is_valid_count(self, comment, history):
        history = pd.concat([history, pd.DataFrame([comment_to_dict(comment)])], ignore_index=True)
        valid_history = self.is_valid_thread(history)[0]
        valid_count = self.looks_like_count(comment)
        valid_user = not counters.is_ignored_counter(str(comment.author))
        return valid_history and valid_count and valid_user, history

    def get_history(self, comment):
        """Fetch enough previous comments to be able to determine whether replies to
        `comment` are valid according to the side thread rules.
        """
        return self.rule.get_history(comment)

    def looks_like_count(self, comment):
        return comment.body in utils.deleted_phrases or self.form(comment.body)

    def set_comment_to_count(self, f):
        self.comment_to_count = f

    def wrapped_comment_to_count(self, comment):
        comment_to_count = (
            self.comment_to_count if self.comment_to_count is not None else base_n_count(10)
        )
        try:
            return comment_to_count(comment)
        except ValueError:
            return np.nan

    def find_errors(self, history, offset=0):
        """Find points in the history of a side thread where an incorrect count was posted.

        Parameters:
          - history: Either a string representing the comment id of
            the leaf comment in the thread to be investigated, or a pandas
            dataframe with (at least) a "body" column that contains the markdown
            string of each comment in the thread.
          - offset: How much the comments in the chain are shifted with respect
            to some platonic "true chain". Used in case some broken chains mean
            that there isn't a linear thread from start to finish.

        Returns:
          - The comments in the history where an uncorrected error was introduced

        In order to do this, we need to use the `comment_to_count` member of
        the side thread to go from the string representation of a comment to
        the corresponding count. This is potentially different for each side
        thread.

        Errors are defined narrowly to avoid having too many false positives. A
        comment is considered to introduce an error if:

          - Its count is not one more than the previous count AND
          - Its count is not two more than the last but one count AND
          - Its count doesn't match where the count should be according to the
            position in the thread.

        The last criterion means that counts which correct previous errrors won't be included.

        Additionally, to avoid rehashing errors which have already been
        corrected, only comments after the last correct count in the thread
        will be considered.

        """
        if isinstance(history, str):
            self.history = pd.DataFrame(tn.fetch_comments(history))
            history = self.history

        counts = history["body"].apply(self.wrapped_comment_to_count)
        # Errors are points where the count doesn't match the index difference
        errors = counts - counts.iloc[0] - offset != counts.index
        # But only errors after the last correct value are interesting
        errors[: errors.where((~errors)).last_valid_index()] = False
        mask = errors & (counts.diff() != 1) & (counts.diff(2) != 2)
        return history[mask]



import logging


def setup(logger, verbose, quiet):
    logger.propagate = False
    logger.handlers.clear()
    logger.setLevel(30 - 10 * (1 - quiet) * (1 + verbose))
    ch = logging.StreamHandler()
    ch.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(ch)



"""
The main entry point for the command line interface.
See the subcommands for details on their behaviour.
"""
import click

from .ftf import pin_or_create_ftf
from .log_thread import log
from .update_thread_directory import update_directory
from .validate import validate
from .weekly_side_thread_stats import generate_stats_post


@click.group(
    commands=[log, validate, update_directory, pin_or_create_ftf, generate_stats_post],
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option()
def cli():
    pass


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



import logging
import os
import sqlite3
from pathlib import Path

import pandas as pd

from rcounting import counters, models, parsing

printer = logging.getLogger(__name__)


class ThreadLogger:
    """Simple class for logging either to a database or to a csv file"""

    def __init__(self, sql, output_directory, filename=None, is_main=True):
        self.sql = sql
        self.is_main = is_main
        assert self.is_main or self.sql, "Only sql output is supported for side threads"
        self.output_directory = output_directory
        self.last_checkpoint = ""
        if self.sql:
            self.setup_sql(filename)
        self.log = self.log_sql if self.sql else self.log_csv

    def setup_sql(self, filename):
        """Connect to the database and get list of existing submissions, if any"""
        last_checkpoint = ""
        known_submissions = []
        if filename is None:
            filename = "counting.sqlite"
        path = self.output_directory / filename
        printer.warning("Writing submissions to sql database at %s", path)
        db = sqlite3.connect(path)
        try:
            known_submissions = pd.read_sql("select * from submissions", db)[
                "submission_id"
            ].tolist()
            checkpoints = pd.read_sql(
                "select last_submission.submission_id "
                "from last_submission join submissions "
                "on last_submission.submission_id = submissions.submission_id "
                "order by timestamp",
                db,
            )
            last_checkpoint = checkpoints.iat[-1, 0]
        except pd.io.sql.DatabaseError:
            pass
        self.db = db
        self.last_checkpoint = last_checkpoint
        self.known_submissions = known_submissions

    def is_already_logged(self, submission):
        """
        Determine whether a submission has already been logged based on its first
        comment. It's important to do it this way because fetching a whole submission
        is expensive, so we want to avoid having to do that.
        """
        if self.sql:
            return submission.id in self.known_submissions
        comment = submission.comments[0]
        body = parsing.strip_markdown_links(comment.body)
        basecount = parsing.find_count_in_text(body)
        hoc_path = self.output_directory / Path(f"{basecount}.csv")
        return os.path.isfile(hoc_path)

    def log_sql(self, comment, df):
        """Save one submission to a database"""
        submission = pd.Series(models.submission_to_dict(comment.submission))
        submission = submission[["submission_id", "username", "timestamp", "title", "body"]]
        if self.is_main:
            submission["base_count"] = base_count(df)
        df.to_sql("comments", self.db, index_label="position", if_exists="append")
        submission.to_frame().T.to_sql("submissions", self.db, index=False, if_exists="append")

    def log_csv(self, comment, df):
        """Save one submission to a csv file"""
        n = base_count(df)
        path = self.output_directory / Path(f"{n}.csv")

        columns = ["username", "timestamp", "comment_id", "submission_id"]
        output_df = df.set_index(df.index + n)[columns].iloc[1:]
        printer.debug("Writing submission log to %s", path)
        header = ["username", "timestamp", "comment_id", "submission_id"]
        with open(path, "w", encoding="utf8") as f:
            print(f"# {comment.submission.title}", file=f)
            print(output_df.to_csv(index_label="count", header=header), file=f, end="")

    def update_checkpoint(self):
        if not self.sql:
            return
        newest_submission = pd.read_sql(
            "select submission_id from submissions order by timestamp", self.db
        ).iloc[-1]
        newest_submission.name = "submission_id"
        newest_submission.to_sql("last_submission", self.db, index=False, if_exists="append")


def update_counters_table(db):
    counting_users = pd.read_sql("select distinct username from comments", db)
    counting_users["canonical_username"] = counting_users["username"].apply(counters.apply_alias)
    counting_users["is_mod"] = counting_users["username"].apply(counters.is_mod)
    counting_users["is_banned"] = counting_users["username"].apply(counters.is_banned_counter)
    counting_users.to_sql("counters", db, index=False, if_exists="replace")


def base_count(df):
    return int(round((df["body"].apply(parsing.wrapped_count_in_text) - df.index).median(), -3))



import itertools
import os

from rcounting.units import HOUR, MINUTE


def flatten(mylist):
    return [element for sublist in mylist for element in sublist]


def partition(mylist, indices):
    return [mylist[i:j] for i, j in zip([0] + indices, indices + [None])]


def is_leap_year(n):
    return n % 4 == 0 and (n % 400 == 0 or n % 100 != 0)


def format_timedelta(timedelta):
    def format_one_interval(n, unit):
        if n == 0:
            return ""
        return f"{n} {unit}{'s' if n > 1 else ''}"

    days = timedelta.days
    hours, rem = divmod(timedelta.seconds, HOUR)
    minutes, seconds = divmod(rem, MINUTE)
    amounts = [days, hours, minutes, seconds]
    units = ["day", "hour", "minute", "second"]
    formatted = [format_one_interval(n, unit) for n, unit in zip(amounts, units)]
    return ", ".join([x for x in formatted if x])


def ensure_directory(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


deleted_phrases = ["[deleted]", "[removed]", "[banned]"]


def chunked(iterable, n):
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)



import datetime as dt


def get_ftf_timestamp():
    current_time = dt.datetime.now(dt.timezone.utc)
    threshold_date = (
        current_time.date()
        - dt.timedelta(days=current_time.weekday())
        + dt.timedelta(days=4, weeks=-1)
    )
    threshold_timestamp = dt.datetime.combine(threshold_date, dt.time(7, tzinfo=dt.timezone.utc))
    if current_time - threshold_timestamp >= dt.timedelta(weeks=1):
        threshold_timestamp += dt.timedelta(weeks=1)
    return threshold_timestamp


def is_within_threshold(post):
    """
    Check if a post was made after the most recent Friday at 0700 UTC
    """
    return post.created_utc >= get_ftf_timestamp().timestamp()



# pylint: disable=import-outside-toplevel
"""Validate the thread ending at COMMENT_ID according to the specified rule."""
import click

rule_dict = {
    "default": "default",
    "wait2": "wait 2",
    "wait3": "wait 3",
    "wait9": "wait 9",
    "wait10": "wait 10",
    "once_per_thread": "once per thread",
    "slow": "slow",
    "slower": "slower",
    "slowestest": "slowestest",
    "only_double_counting": "only double counting",
    "fast_or_slow": "fast or slow",
    "no_repeating": "no repeating digits",
    "not_any": "not any of those",
}


@click.command(no_args_is_help=True)
@click.option(
    "--rule",
    help="Which rule to apply. Default is no double counting",
    default="default",
    type=click.Choice(list(rule_dict.keys()), case_sensitive=False),
)
@click.argument("comment_id")
def validate(comment_id, rule):
    """Validate the thread ending at COMMENT_ID according to the specified rule."""
    import pandas as pd

    from rcounting import side_threads as st
    from rcounting import thread_navigation as tn
    from rcounting.reddit_interface import reddit

    comment = reddit.comment(comment_id)
    print(f"Validating thread: '{comment.submission.title}' according to rule {rule}")
    comments = pd.DataFrame(tn.fetch_comments(comment))
    side_thread = st.get_side_thread(rule_dict[rule])
    result = side_thread.is_valid_thread(comments)
    if result[0]:
        print("All counts were valid")
        errors = side_thread.find_errors(comments)
        if errors.empty:
            print("The last count in the the thread has the correct value")
        else:
            print(
                f"The last count in the thread has an incorrect value. "
                f"Earlier errors can be found at {errors}"
            )
    else:
        print(f"Invalid count found at http://reddit.com{reddit.comment(result[1]).permalink}!")


if __name__ == "__main__":
    validate()  # pylint: disable=no-value-for-parameter



import string
from typing import Iterable

from rcounting import parsing

alphanumeric = string.digits + string.ascii_lowercase


def validate_from_tokens(valid_tokens: str | Iterable[str], strip_links=True):
    def looks_like_count(comment_body: str) -> bool:
        body = comment_body.lower()
        if strip_links:
            body = parsing.strip_markdown_links(body)
        return any(character.lower() in body for character in valid_tokens)

    return looks_like_count


def base_n(n=10, strip_links=True):
    return validate_from_tokens(alphanumeric[:n], strip_links)


def permissive(_: str) -> bool:
    return True



"""A collection of functions for parsing texts and extracting urls and counts"""
import re

from rcounting.reddit_interface import extract_from_short_link


def find_body(post):
    return post.body if hasattr(post, "body") else post.selftext


def extract_count_string(body: str, base: int = 10):
    """
    Extract a normalized base n representation of an integer from a messy comment.
    Try to account for various strategies for separating thousands digits.
    """
    characters = "0123456789abcdefghijklmnopqrstuvwxyz"[:base]
    separators = "' ,.*/"  # non-whitespace separators people have used
    first_line = "".join(normalize_comment_body(body).split()).translate(
        str.maketrans("", "", separators)
    )
    match = re.search(f"[{characters}]+", first_line.lower())
    if match is not None:
        return match.group()
    raise ValueError(f"Unable to extract count in base {base} from comment body: {body}")


def find_count_in_text(body: str, base: int = 10):
    """Extract a base n integeger from a string, relying on `extract_count_string`"""
    return int(extract_count_string(body, base), base)


def wrapped_count_in_text(body: str, base: int = 10):
    try:
        return find_count_in_text(body, base)
    except ValueError:
        return float("nan")


def find_urls_in_text(body):
    """Extract all substrings of a string that look like '/comments/id/stuff/id'

    Returns a list of (submission_id, comment_id) tuples; the `comment_id`
    field is potentially blank, if someone accidentally linked to just a
    submission rather than a comment.
    """
    # reddit lets you link to comments and posts with just /comments/stuff,
    # so everything before that is optional. We only capture the bit after
    # r/counting/comments, since that's what has the information we are
    # interested in.
    url_regex = r"/comments/([\w]+)(?:/[^/]*/([\w]*)|)"
    urls = re.findall(url_regex, body)

    # reddit has introduced new short links which are completely opaque to the
    # api. We need to handle those separately.
    new_url_regex = "reddit.com/r/counting/s/([A-Za-z0-9]+)"
    new_url_prefix = "https://www.reddit.com/r/counting/s/"
    short_links = [new_url_prefix + x for x in re.findall(new_url_regex, body)]
    extra_urls = [extract_from_short_link(l) for l in short_links]
    return urls + extra_urls


def post_to_count(reddit_post):
    """Extract an integer from a reddit submission or comment"""
    return find_count_in_text(find_body(reddit_post))


def post_to_urls(reddit_post):
    """Find urls in a reddit submission or comment"""
    return find_urls_in_text(find_body(reddit_post))


def parse_markdown_links(body):
    """
    Find markdown links of the form [description](link) in a string.
    Correctly handling escaped backslashes in the link portion
    """
    regex = r"\[(.*?)\]\((.+?(?<!\\))\)"
    links = re.findall(regex, body)
    return links


def strip_markdown_links(body):
    """Replace all markdown links of the form [description](link) with description."""
    regex = r"\[(.*?)\]\((.+?(?<!\\))\)"
    replacement = r"\1"
    return re.sub(regex, replacement, body)


def body_from_title(title):
    return "|".join(title.split("|")[1:])


def normalize_comment_body(comment):
    first_line = comment.split("\n")[0]
    no_links = strip_markdown_links(first_line)
    return no_links.strip()


def parse_directory_page(directory_page):
    """Tag each paragraph of a directory page by whether it represents text or a table."""
    directory_page = re.sub(r"\n{2,}", r"\n\n", directory_page)
    regex = r"^.*\|.*\|.*$"
    tagged_results = []
    text = []
    rows = []
    for line in directory_page.split("\n"):
        if bool(re.match(regex, line)):
            # flush the text buffer
            if text:
                tagged_results.append(["text", "\n".join(text).strip("\n")])
                text = []
            rows.append(line)
        else:
            # flush the table buffer
            if rows and len(rows) > 2:
                rows = [parse_row(row) for row in rows[2:]]
                tagged_results.append(["table", rows])
                rows = []
            elif rows:
                # We have an incorrectly formatted table - it's only two lines
                # long! We'll treat it as text and not update it
                text = rows
                rows = []
            text.append(line)
    if text:
        tagged_results.append(["text", "\n".join(text)])
    if rows:
        rows = [parse_row(row) for row in rows[2:]]
        tagged_results.append(["table", rows])
    return tagged_results


def parse_row(markdown_row):
    """Extract the side thread attributes from a row in a markdown table"""
    first, current, count = markdown_row.split("|")
    name, first_submission = parse_markdown_links(first)[0]
    name = name.strip()
    first_submission_id = first_submission.split("/")[-1]
    title, link = parse_markdown_links(current)[0]
    title = title.strip()
    submission_id, comment_id = find_urls_in_text(link)[0]
    comment_id = None if not comment_id else comment_id
    count = count.strip()
    return name, first_submission_id, title, submission_id, comment_id, count


def find_urls_in_submission(submission):
    """Extract urls from both the body of the submission and the top-level comments"""
    # Get everything that looks like a url in the body of the post
    yield from find_urls_in_text(submission.selftext)
    # And then in every top-level comment
    for comment in submission.comments:
        yield from find_urls_in_text(comment.body)


def is_revived(title):
    """Determine whether the title indicates that a submission is a revival"""
    regex = r"\(*reviv\w*\)*"
    return re.search(regex, title.lower())


def name_sort(name):
    """Intelligent sorting for strings mixed with digits"""
    title = name.translate(str.maketrans("", "", "'\"()^/*")).lower()
    return tuple(int(c) if c.isdigit() else c for c in re.split(r"(\d+)", title))


def normalise_title(title):
    """
    Normalise a string for posting to the directory

    - Literal pipes are replaced by an equivalent character, to avoid messing up tables
    - The format for indicating revivals is standardised.
    """
    title = title.translate(str.maketrans("[]", "()"))
    title = title.replace("|", "&#124;")
    revived = is_revived(title)
    if revived:
        start, end = revived.span()
        return title[:start] + "(Revival)" + title[end:]
    return title



# pylint: disable=import-outside-toplevel
import datetime as dt
import logging
import sqlite3
import time
from pathlib import Path

import click
import pandas as pd
from fuzzywuzzy import fuzz
from prawcore.exceptions import TooManyRequests

from rcounting import configure_logging, counters, ftf, models, parsing
from rcounting import thread_directory as td
from rcounting import thread_navigation as tn
from rcounting import units

printer = logging.getLogger("rcounting")

WEEK = 7 * units.DAY


def find_directory_revision(subreddit, threshold):
    """Find the first directory revision which was made after a threshold
    timestamp. If no such revision exists, return the latest revision.

    """
    revisions = subreddit.wiki["directory"].revisions()
    old_revision = next(revisions)
    first_revision = old_revision
    for new_revision in revisions:
        if new_revision["timestamp"] < threshold:
            return old_revision
        old_revision = new_revision
    return first_revision


def get_directory_counts(reddit, directory, ftf_timestamp, db):
    threshold = ftf_timestamp - WEEK
    multiple = 1
    try:
        threads = pd.read_sql("select * from threads", db).set_index("thread_id")
    except pd.io.sql.DatabaseError:
        threads = pd.DataFrame()

    for row in directory.rows[1:]:
        try:
            checkpoint_timestamp = threads.loc[row.first_submission, "checkpoint_timestamp"]
        except KeyError:
            checkpoint_timestamp = 0
        if checkpoint_timestamp == threshold:
            printer.info("row %s has already been logged this week!", row.name)
        else:
            printer.info("Getting history for %s", row.name)
            while True:
                try:
                    submissions, comments = get_side_thread_counts(reddit, row, threshold)
                    multiple = 1
                    break
                except TooManyRequests:
                    time.sleep(30 * multiple)
                    multiple *= 1.5
            df = pd.DataFrame([models.comment_to_dict(c) for c in comments])
            if not df.empty:
                df = df[df["timestamp"] < ftf_timestamp]
                df.to_sql("comments", db, if_exists="append", index=False)
            write_submissions(db, submissions, row.first_submission)
            update_checkpoint(db, row.first_submission, row.name, threshold)
    query = (
        "SELECT comments.*, submissions.thread_id "
        "FROM comments JOIN submissions ON comments.submission_id==submissions.submission_id "
        f"WHERE comments.timestamp >= {threshold}"
    )
    comments = pd.read_sql(query, db).drop_duplicates()
    threads = pd.read_sql("select * from threads", db)

    return pd.merge(comments, threads, left_on="thread_id", right_on="thread_id", how="left")


def update_checkpoint(db, thread_id, name, timestamp):
    row = {"thread_id": thread_id, "thread_name": name, "checkpoint_timestamp": timestamp}
    try:
        threads = pd.read_sql("select * from threads", db)
        if (threads["thread_id"] == thread_id).any():
            threads.loc[threads["thread_id"] == thread_id, "checkpoint_timestamp"] = timestamp
        else:
            threads = pd.concat([threads, pd.DataFrame([row])])
    except pd.io.sql.DatabaseError:
        threads = pd.DataFrame([row])

    threads.to_sql("threads", db, index=False, if_exists="replace")


def write_submissions(db, submissions, thread_id):
    df = pd.DataFrame([models.submission_to_dict(submission) for submission in submissions])
    df["thread_id"] = thread_id
    try:
        existing_submissions = pd.read_sql("select * from submissions", db)
    except pd.io.sql.DatabaseError:
        existing_submissions = pd.DataFrame()

    new_submissions = pd.concat([existing_submissions, df]).drop_duplicates()
    new_submissions.to_sql("submissions", db, index=False, if_exists="replace")


def get_side_thread_counts(reddit, row, threshold):
    """Return a list of all counts in the side thread represented by row that
    occurred after `threshold`, as well as the corresponding submissions the
    comments belong to.

    """
    submission_id = row.submission_id
    comment_id = row.comment_id
    submission = reddit.submission(submission_id)
    comments = []
    submissions = [submission]
    while submission.created_utc >= threshold:
        printer.debug("Fetching submission %s", submission.title)
        tree = models.CommentTree(reddit=reddit, get_missing_replies=False)
        new_comments = tree.walk_up_tree(comment_id)
        if new_comments is not None:
            comments += new_comments[:-1]
        try:
            submission_id, comment_id = tn.find_previous_submission(submission)
            if not comment_id:
                return submissions, comments
        # We've hit the first submission in a new side thread
        except StopIteration:
            return submissions, comments
        submission = reddit.submission(submission_id)
        submissions.append(submission)
    tree = models.CommentTree(reddit=reddit, get_missing_replies=False)
    new_comments = tree.walk_up_tree(comment_id, cutoff=threshold)
    if new_comments is not None:
        comments += new_comments
    return submissions, comments


def get_weekly_stats(reddit, subreddit, ftf_timestamp, filename):
    db = sqlite3.connect(filename)

    revision = find_directory_revision(subreddit, ftf_timestamp)
    contents = revision["page"].content_md.replace("\r\n", "\n")
    directory = td.Directory(parsing.parse_directory_page(contents), "directory")
    return get_directory_counts(reddit, directory, ftf_timestamp, db)


def pprint(date):
    return date.strftime("%A %B %d, %Y")


def stats_post(stats, ftf_timestamp):
    end = dt.date.fromtimestamp(ftf_timestamp)
    start = dt.date.fromtimestamp(ftf_timestamp - WEEK)
    stats["canonical_username"] = stats["username"].apply(counters.apply_alias)

    top_counters = (
        stats.groupby("canonical_username")
        .size()
        .sort_values(ascending=False)
        .to_frame()
        .reset_index()
    )

    top_counters = top_counters[
        ~top_counters["canonical_username"].apply(counters.is_banned_counter)
    ].reset_index(drop=True)

    top_counters.index += 1
    tc = top_counters["canonical_username"].to_numpy()

    top_threads = (
        stats.groupby("thread_name").size().sort_values(ascending=False).to_frame().reset_index()
    )
    top_threads.index += 1
    s = (
        f"Weekly side thread stats from {pprint(start)} to {pprint(end)}. "
        f"Congratulations to u/{tc[0]}, u/{tc[1]}, and u/{tc[2]}!\n\n"
    )
    s += f"Total weekly side thread counts: **{len(stats)}**\n\n"
    s += "Top 15 side thread counters:\n\n"
    s += top_counters.head(15).to_markdown(headers=["**Rank**", "**User**", "**Counts**"])
    s += "\n\nTop 5 side threads:\n\n"
    s += top_threads.head(5).to_markdown(headers=["**Rank**", "**Thread**", "**Counts**"])
    s += "\n\n----\n\n"
    s += "*This comment was made by a script; check it out "
    # pylint: disable-next=line-too-long
    s += "[here](https://github.com/cutonbuminband/rcounting/blob/main/rcounting/scripts/weekly_side_thread_stats.py)*"
    return s


def is_duplicate(body, post):
    scores = [(comment.permalink, fuzz.ratio(body, comment.body)) for comment in post.comments]
    if not scores:
        return False, None
    link, score = max(scores, key=lambda x: x[1])
    return score > 90, link


@click.command(name="st-stats")
@click.option(
    "--dry-run", is_flag=True, help="Write results to console instead of making a comment"
)
@click.option("--verbose", "-v", count=True, help="Print more output")
@click.option("--quiet", "-q", is_flag=True, default=False, help="Suppress output")
@click.option(
    "--filename",
    "-f",
    type=click.Path(path_type=Path),
    help=("What file to write to. If none is specified, side_threads.sqlite is used"),
)
def generate_stats_post(filename, dry_run, verbose, quiet):
    """Load all the side thread counts made in the previous FTF period and
    store them in a database. Then post a side thread stats comment in the
    current FTF as long as:

    - The current FTF is not stale
    - The comment has not been posted before.

    """
    t_start = dt.datetime.now()
    ftf_timestamp = ftf.get_ftf_timestamp().timestamp()

    from rcounting.reddit_interface import reddit, subreddit

    configure_logging.setup(printer, verbose, quiet)
    stats = get_weekly_stats(reddit, subreddit, ftf_timestamp, filename)
    body = stats_post(stats, ftf_timestamp)
    if dry_run:
        print(body)
    else:
        ftf_post = subreddit.sticky(number=2)
        duplicate, link = is_duplicate(body, ftf_post)
        if ftf.is_within_threshold(ftf_post) and not duplicate:
            ftf_post.reply(body)
        elif not duplicate:
            printer.warning("Not posting stats comment. Pinned FTF is stale")
        else:
            s = "Not posting stats comment. Existing comment found at https://www.reddit.com%s"
            printer.warning(s, link)
    printer.warning("Running the script took %s", dt.datetime.now() - t_start)


if __name__ == "__main__":
    generate_stats_post()  # pylint: disable=no-value-for-parameter



"""
A collection of functions for analysing counting data.
See also https://github.com/cutonbuminband/rcounting/blob/main/doc/examples.org
for examples of working with the data.
"""
from pathlib import Path

import numpy as np
import pandas as pd
from numpy import pi
from numpy.fft import fftshift, irfft, rfft
from scipy.special import i0

from rcounting import counters, utils
from rcounting.counters.counters import is_banned_counter
from rcounting.units import DAY


def load_csvs(start, n, directory="."):
    """Load n csv files into a single dataframe"""
    results = []
    for _ in range(n):
        path = Path(directory) / Path(f"{start}.csv")
        df = pd.read_csv(path, comment="#")
        if len(df) % 1000 != 0:
            print(path)
        df = df.set_index("comment_id")
        results.append(df)
        start += 1000
    return pd.concat(results)


def hoc_string(df, title):
    """
    Calculate the thread participation table of a data frame.
    Return a string representation of it.
    """
    getter = counters.apply_alias(df.iloc[-1]["username"])

    def hoc_format(username):
        username = counters.apply_alias(username)
        return f"**/u/{username}**" if username == getter else f"/u/{username}"

    df["hoc_username"] = df["username"].apply(hoc_format)
    dt = pd.to_timedelta(df.iloc[-1].timestamp - df.iloc[0].timestamp, unit="s")
    table = df.iloc[1:]["hoc_username"].value_counts().to_frame().reset_index()
    data = table.set_index(table.index + 1).to_csv(None, sep="|", header=0)

    header = f"Thread Participation Chart for {title}\n\nRank|Username|Counts\n---|---|---"
    footer = (
        f"It took {len(table)} counters {utils.format_timedelta(dt)} to complete this thread."
        f" Bold is the user with the get\ntotal counts in this chain logged: {len(df) - 1}"
    )
    return "\n".join([header, data, footer])


def response_graph(df, n=250, username_column="username"):
    """
    Calculate the network graph of the top n counters.
    Create a directed edge a->b in the graph if a has ever replied to b.
    Weight each edge by the number of replies
    """
    indices = df.groupby(username_column).size().sort_values(ascending=False).index
    indices = [x for x in indices if not is_banned_counter(x)][:n]
    edges = df[username_column].isin(indices) & df[username_column].shift(1).isin(indices)
    top = pd.concat([df[username_column], df[username_column].shift()], axis=1).loc[edges]
    top.columns = ["username", "replying_to"]
    graph = top.groupby(["username", "replying_to"], as_index=False).size()
    graph.columns = ["Source", "Target", "Weight"]
    return graph


def effective_number_of_counters(c):
    """
    Calculate the effective number of parties for a given reply distribution.
    See also https://en.wikipedia.org/wiki/Effective_number_of_parties
    """
    normalised_counters = c / c.sum()
    return 1 / (normalised_counters ** 2).sum()


def capture_the_flag_score(submission):
    """Calculate how long each user has had the latest count"""
    submission["score"] = submission["timestamp"].diff().shift(-1)
    return submission.groupby("username")["score"].sum()


def vonmises_distribution(x, mu=0, kappa=1):
    """Calculate the von mises distribution; the gaussian on a circle"""
    return np.exp(kappa * np.cos(x - mu)) / (2.0 * np.pi * i0(kappa))


def normal_distribution(x, mu=0, sigma=1):
    """Calculate the gaussian distribution with mean mu and standard deviation sigma"""
    return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / (sigma * np.sqrt(np.pi * 2))


def fft_kde_on_unit_circle(data, n_bins, kernel=vonmises_distribution, **params):
    """The fft approximation to the kde of data on the unit circle

    params:
    data: array-like: The data points to be approximated. Assumed to be distributed on [-pi, pi]
    n_bins: number of bins to use for approximation
    kernel: The kernel used to approximate the data. Default is the gaussian on the circle
    **params: parameters apart from x needed to define the kernel.

    returns:
    x: array-like: points on the unit circle at which to evaluate the density
    kde: Estimated density, normalised so that ∫kde dx = 1.

    In genergal, kernel density estimation means replacing each discrete point
    with some normalised distribution centered at that point, and then adding
    up all these results. Mathematically, this corresponds to convolving the
    original data with the chosen distribution. In fourier space, a convolution
    transforms to a product, so for periodic data we can get a fast
    approximation to this by finding the Fourier transform of our data, the
    Fourier transform of our kernel, multiplying them together and finding the
    inverse transform.

    """

    x_axis = np.linspace(-pi, pi, n_bins + 1, endpoint=True)
    hist, edges = np.histogram(data, bins=x_axis)
    x = np.mean([edges[1:], edges[:-1]], axis=0)
    kernel = kernel(x=x, **params)
    kde = fftshift(irfft(rfft(kernel) * rfft(hist)))
    kde /= np.trapz(kde, x=x)
    return x, kde


def fft_kde(data, n_bins, kernel=vonmises_distribution, **params):
    """Prepare time data for use with `fft_kde_on_unit_circle`"""
    minval = 0
    maxval = DAY
    if kernel == "vonmises_distribution":
        kernel = vonmises_distribution
    elif kernel == "normal_distribution":
        kernel = normal_distribution
    data_on_unit_circle = (data - minval) / (maxval - minval) * 2 * pi - pi
    x, kde = fft_kde_on_unit_circle(data_on_unit_circle, n_bins, kernel, **params)
    x = (x + pi) / (2 * pi) * (maxval - minval) + minval
    kde /= np.trapz(kde, x=x)
    return x, kde


def even_odd_counts(df, n=50):
    indices = df.groupby("username").size().sort_values(ascending=False).index
    top_counters = [x for x in indices if not counters.is_banned_counter(x)][:n]
    df["is_even"] = df["position"] % 2 == 0
    offsets = [
        "1gm10t",
        "7hn2tm",
        "b471wg",
        "bz6r0g",
        "d6pgni",
        "ebnh39",
        "grggc0",
        "oj50hj",
        "ob4a2h",
        "t81gug",
    ]
    for offset in offsets:
        df.loc[df["submission_id"] == offset, "is_even"] = (
            1 - df.loc[df["submission_id"] == offset, "is_even"]
        )
    df["is_odd"] = 1 - df["is_even"]
    subset = df.query("username in @top_counters")
    table = subset[["username", "is_even", "is_odd"]].groupby("username").sum()
    table.columns = ["n_even", "n_odd"]
    table["difference"] = table["n_odd"] - table["n_even"]
    table["relative_difference"] = table["difference"] / (table["n_even"] + table["n_odd"]) * 100
    table["absolute_difference"] = abs(table["relative_difference"])
    return table



import bisect

import networkx as nx
import numpy as np

# See e.g. https://iopscience.iop.org/article/10.1088/1367-2630/14/8/083030


def prepare_graph(graph):
    G = nx.DiGraph(graph)
    G.remove_edges_from(nx.selfloop_edges(G))

    G = symmetrize(G)
    scale_weights(G)
    return G


def weighted_product(x, y, p):
    """
    Calculate the weighted product of x and y with scaling factor p:

    x^p * y^(1 - p)

    If p = 0, only the first factor counts; if p = 1 only the second. Values in between are a mix.
    """
    assert 0 <= p <= 1
    q = 1 - p
    return x ** q * y ** p


def calculate_weighted_degrees(graph, p=0.5):
    degrees = graph.degree()
    weighted_degrees = graph.degree(weight="weight")
    return {node: weighted_product(degrees[node], weighted_degrees[node], p) for node in graph}


def weighted_degree(edgelist, p=0.5):
    return weighted_product(len(edgelist), sum(edgelist.values()), p)


def weighted_core_number(graph, p=0.5):
    degrees = calculate_weighted_degrees(graph, p)
    nodes = sorted(degrees, key=degrees.get)
    weights = [degrees[node] for node in nodes]
    neighbors = {v: dict((x[1:] for x in graph.edges(v, data="weight"))) for v in graph}
    core = degrees
    max_weight = 0
    for _ in range(len(nodes)):
        w = weights.pop(0)
        if w > max_weight:
            max_weight = w
        w = max_weight
        v = nodes.pop(0)
        core[v] = w
        for u in neighbors[v].keys():
            if core[u] > core[v]:
                del neighbors[u][v]
                old_position = bisect.bisect(weights, core[u]) - 1
                new_core = weighted_degree(neighbors[u], p)
                core[u] = new_core
                new_position = bisect.bisect(weights, core[u])
                nodes.remove(u)
                nodes.insert(new_position, u)
                del weights[old_position]
                weights.insert(new_position, core[u])
    return core


def symmetrize(graph):
    UG = graph.to_undirected()
    for node in graph:
        for successor in graph.successors(node):
            if successor in graph.predecessors(node):
                UG.edges[node, successor]["weight"] = (
                    graph.edges[node, successor]["weight"] + graph.edges[successor, node]["weight"]
                )
    return UG


def scale_weights(graph, scaling=lambda x: 1 + np.log10(x)):
    for node in graph:
        for neighbor in graph.neighbors(node):
            graph.edges[node, neighbor]["weight"] = scaling(graph.edges[node, neighbor]["weight"])
    return graph


# Next steps:
# * The onion decomposition
# https://www.nature.com/articles/srep31708
#
# Idea: make a k-shell decomposition (recursively remove nodes of degree <= k
# from the graph until there are no more, then do the same for k+1), but keep
# track of which/how nodes are removed for each pass *within* a shell, and not
# just what the k value is.



import datetime
import logging

from praw.exceptions import DuplicateReplaceException

from rcounting import models, parsing, reddit_interface

printer = logging.getLogger(__name__)


def find_previous_submission(submission):
    """
    Find the previous reddit submission in the chain of counts. The code
    will look first in the body of the submission text and then in the top
    level comments for everything that looks like a link to a reddit comment.
    It will take the first comment link it finds and use that as the previous
    submission in the chain. Failing that, it will take the first link to a
    reddit submission it finds and use as the previous link in the chain.
    Failing that, it will just fail.

    """

    urls = filter(
        lambda x: int(x[0], 36) < int(submission.id, 36),
        parsing.find_urls_in_submission(submission),
    )
    next_submission_id = None
    new_submission_id, new_get_id = next(urls)
    # This gets a bit silly but if the first link we found only had a
    # submission id and not a comment id we scan through all the rest, and then
    # take the first one that also has a comment id. If none of them do, we
    # fall back to the first link we found
    while not new_get_id:
        try:
            next_submission_id, new_get_id = next(urls)
        except StopIteration:
            break
    if next_submission_id is not None and new_get_id:
        return next_submission_id, new_get_id
    return new_submission_id, new_get_id


def find_get_in_submission(submission_id, get_id, validate_get=True):
    reddit = reddit_interface.reddit
    if not get_id:
        get_id = find_deepest_comment(submission_id, reddit)
    comment = reddit.comment(get_id)
    if validate_get:
        get = find_get_from_comment(comment)
    else:
        get = reddit.comment(get_id)
    printer.debug(
        "Found previous get at: http://reddit.com/comments/%s/_/%s/",
        get.submission,
        get.id,
    )
    return get


def find_previous_get(submission, validate_get=True):
    """
    Find the get of the previous reddit submission in the chain of counts.

    There's a user-enforced convention that the previous get should be linked
    in either the body of the submission, or the first comment.

    Usually this convention is followed, but sometimes this isn't done.
    Frequently, even in those cases the get will be linked in a different
    top-level comment.

    Parameters:

    submission: A reddit submission instance for which we want to find the parent
    validate_get: Whether or not the prorgram should check that the linked comment ends in 000,
    and if it doesn't, try to find a nearby comment that does.
    """
    reddit = reddit_interface.reddit
    submission = submission if hasattr(submission, "id") else reddit.submission(submission)
    new_submission_id, new_get_id = find_previous_submission(submission)
    return find_get_in_submission(new_submission_id, new_get_id, validate_get)


def find_deepest_comment(submission, reddit):
    """
    Find the deepest comment on a submission
    """
    if not hasattr(submission, "id"):
        submission = reddit.submission(submission)
        submission.comment_sort = "old"
    comments = models.CommentTree(reddit=reddit)
    for comment in submission.comments:
        comments.add_missing_replies(comment)
    return comments.deepest_node.id


def search_up_from_gz(comment, max_retries=5):
    """Look for a count up to max_retries above the linked_comment"""
    for i in range(max_retries):
        try:
            count = parsing.post_to_count(comment)
            return count, comment
        except ValueError:
            if i == max_retries:
                raise
            comment = comment.parent()
    raise ValueError(f"Unable to find count in {comment.submission.permalink}")


def find_get_from_comment(comment):
    """Look for the get either above or below the linked comment"""
    count, comment = search_up_from_gz(comment)
    comment.refresh()
    replies = comment.replies
    try:
        replies.replace_more(limit=None)
    except DuplicateReplaceException:
        pass
    while count % 1000 != 0:
        comment = comment.replies[0]
        count = parsing.post_to_count(comment)
    return comment


def fetch_comments(comment, limit=None):
    """
    Fetch a chain of comments from root to the supplied leaf comment.
    """
    reddit = reddit_interface.reddit
    tree = models.CommentTree([], reddit=reddit)
    comment_id = getattr(comment, "id", comment)
    comments = tree.comment(comment_id).walk_up_tree(limit=limit)[::-1]
    return [models.comment_to_dict(x) for x in comments]


def fetch_counting_history(subreddit, time_limit):
    """
    Fetch all submissions made to r/counting within time_limit days
    """
    now = datetime.datetime.utcnow()
    submissions = subreddit.new(limit=1000)
    tree = {}
    submissions_dict = {}
    new_submissions = []
    for count, submission in enumerate(submissions):
        submission.comment_sort = "old"
        if count % 20 == 0:
            printer.debug("Processing reddit submission %s", submission.id)
        title = submission.title.lower()
        author = y.name.lower() if (y := submission.author) is not None else None
        if "tidbits" in title or "free talk friday" in title or author == "rcounting":
            continue
        submissions_dict[submission.id] = submission
        urls = parsing.find_urls_in_submission(submission)
        try:
            url = next(filter(lambda x, s=submission: int(x[0], 36) < int(s.id, 36), urls))
            tree[submission.id] = url[0]
        except StopIteration:
            new_submissions.append(submission)
        post_time = datetime.datetime.utcfromtimestamp(submission.created_utc)
        if now - post_time > time_limit:
            break
    else:  # no break
        printer.warning(
            "Threads between %s and %s have not been collected", now - time_limit, post_time
        )

    return (
        models.SubmissionTree(submissions_dict, tree, reddit_interface.reddit),
        new_submissions,
    )



import configparser
import functools
import logging
import math
import os
import re
import string

from fuzzywuzzy import fuzz
from scipy.special import binom

from rcounting import parsing
from rcounting import thread_navigation as tn
from rcounting import utils
from rcounting.units import DAY, HOUR, MINUTE

from .dfa import dfa_threads
from .rules import CountingRule, FastOrSlow, OnlyDoubleCounting
from .side_threads import SideThread
from .validate_count import base_n_count, by_ns_count, count_from_word_list
from .validate_form import base_n, validate_from_tokens

module_dir = os.path.dirname(__file__)
printer = logging.getLogger(__name__)

base_10 = base_n(10)
balanced_ternary = validate_from_tokens("T-0+")
brainfuck = validate_from_tokens("><+-.,[]")
roman_numeral = validate_from_tokens("IVXLCDMↁↂↇ")
mayan_form = validate_from_tokens("Ø1234|-")
twitter_form = validate_from_tokens("@")
parentheses_form = validate_from_tokens("()")


def d20_form(comment_body):
    return "|" in comment_body and base_10(comment_body)


def reddit_username_form(comment_body):
    return "u/" in comment_body


def throwaway_form(comment_body):
    return (fuzz.partial_ratio("u/throwaway", comment_body) > 80) and base_10(comment_body)


def illion_form(comment_body):
    return fuzz.partial_ratio("illion", comment_body) > 80


isenary = {"they're": 1, "taking": 2, "the": 3, "hobbits": 4, "to": 5, "isengard": 0, "gard": 0}
isenary_form = validate_from_tokens(list(isenary.keys()))
isenary_count = functools.partial(count_from_word_list, alphabet=isenary, ignored_chars="!>")

planets = ["mercury", "venus", "earth", "mars", "jupiter", "saturn", "uranus", "neptune"]
planetary_form = validate_from_tokens(planets)
planetary_count = functools.partial(count_from_word_list, alphabet=planets)

colors = ["red", "orange", "yellow", "green", "blue", "indigo", "violet"]
rainbow_form = validate_from_tokens(colors)
rainbow_count = functools.partial(count_from_word_list, alphabet=colors)

with open(os.path.join(module_dir, "us_states.txt"), encoding="utf8") as f:
    us_states = [x.strip().lower() for x in f.readlines()]
us_states_form = validate_from_tokens(us_states)
us_states_count = functools.partial(count_from_word_list, alphabet=us_states, bijective=True)

with open(os.path.join(module_dir, "elements.txt"), encoding="utf8") as f:
    elements = [x.strip() for x in f.readlines()]
element_form = validate_from_tokens(elements)


def element_tokenize(comment_body, _):
    return re.findall("[A-Z][^A-Z]*", comment_body.split("\n")[0])


element_count = functools.partial(
    count_from_word_list, alphabet=elements, tokenize=element_tokenize, bijective=True
)


def permutation_order(word, alphabet, ordered=False, no_leading_zeros=False):
    word_length = len(word)
    if word_length == 0:
        return 0
    index = alphabet.index(word[0])
    position = index - int(no_leading_zeros)
    n_digits = len(alphabet)
    prefix = [] if ordered else alphabet[:index]
    new_alphabet = prefix + alphabet[index + 1 :]
    if ordered:
        first_place_counts = sum(
            math.comb(n_digits - 1 - i, word_length - 1) for i in range(position)
        )
    else:
        first_place_counts = position * math.perm(n_digits - 1, word_length - 1)
    return first_place_counts + permutation_order(word[1:], new_alphabet, ordered=ordered)


def _permutation_count(comment_body, alphabet) -> int:
    alphabet = alphabet.lower()
    word = "".join(
        x for x in parsing.normalize_comment_body(comment_body).lower() if x in alphabet
    )
    l = len(word)
    shorter_words = sum(math.factorial(i) for i in range(1, l))
    return shorter_words + permutation_order(word, alphabet[:l]) - 1


permutation_count = functools.partial(_permutation_count, alphabet="123456789")
letter_permutation_count = functools.partial(_permutation_count, alphabet=string.ascii_lowercase)


def bcd_count(comment):
    count = parsing.extract_count_string(comment, base=2)
    digits = [str(int("".join(y for y in x), 2)) for x in utils.chunked(count, 4)]
    return int("".join(digits))


def nrd_count(comment):
    normalized_comment = parsing.extract_count_string(comment)
    result = 9 * sum(math.perm(9, i - 1) for i in range(1, len(normalized_comment)))
    return result + permutation_order(normalized_comment, string.digits, no_leading_zeros=True)


def nrl_count(comment):
    line = "".join(
        x for x in parsing.normalize_comment_body(comment).lower() if x in string.ascii_lowercase
    )
    shorter_words = sum(math.perm(26, i) for i in range(1, len(line)))
    return shorter_words + permutation_order(line, string.ascii_lowercase)


def powerball_count(comment):
    balls, powerball = parsing.normalize_comment_body(comment).split("+")
    balls = balls.split()
    alphabet = [str(x) for x in range(1, 70)]
    return permutation_order(balls, alphabet, ordered=True) * 26 + int(powerball) - 1


u_squares = [11035, 65039, 129003, 129002, 128998, 129001, 129000, 128999, 128997, 11036]
colored_squares_form = validate_from_tokens([chr(x) for x in u_squares])


@functools.cache
def collatz(n):
    if n == 1:
        return 1
    if n % 2 == 0:
        return 1 + collatz(n // 2)
    return 1 + collatz(3 * n + 1)


def collatz_count(comment):
    regex = r".*\((\d+).*(\d+)\)"
    current, steps = map(int, re.search(regex, comment).groups())
    return sum(collatz(i) for i in range(1, current)) + steps


def ordered_pairs_count(comment_body):
    # A set of brackets containing two integers, separated by at least one non-integer
    regex = r"\(([0-9]+)[^0-9]+([0-9]+)\)"
    x, y = map(int, re.search(regex, comment_body).groups())
    return x**2 + y if y <= x else y**2 + 2 * y - x


def rgb_count(comment_body: str):
    first_line = parsing.normalize_comment_body(comment_body)
    values = map(int, re.findall(r"\d+", first_line))
    return functools.reduce(lambda x, y: 256 * x + y, values)


# an int, then a bracketed int, maybe with a plus or a minus after it
wave_regex = r"(-?\d+).*\((\d+)[\+-]?\)"
double_wave_regex = r"(-?\d+).*\((\d+)\).*\((\d+)\)"


def wave_count(comment):
    comment = parsing.normalize_comment_body(comment)
    match = re.search(wave_regex, comment)
    a, b = [int(x) for x in match.groups()]
    return 2 * b**2 - a


def increasing_type_count(n):
    regex = r"(-?\d+)" + r".*\((\d+)\)" * n

    def count(comment):
        comment = parsing.normalize_comment_body(comment)
        total = 0
        values = [int(x) for x in re.search(regex, comment).groups()]
        for ix, value in enumerate(values):
            total += triangle_n_dimension(ix + 1, value)
        return total

    return count


def triangle_n_dimension(n, value):
    if value == 1:
        return 0
    return math.comb(value - 2 + n, n)


def gaussian_integer_count(comment):
    digits = parsing.extract_count_string(comment, base=4)
    corner = sum((-4) ** ix * int(digit) for ix, digit in enumerate(digits[::-2]))
    return (2 * corner + 1) ** 2


def get_base_powers(n, b):
    r = []
    while n:
        r.append(n % b)
        n = n // b
    return r[::-1]


def palindrome_count(comment_body: str, b=10) -> int:
    count = parsing.find_count_in_text(comment_body, b)
    digits = get_base_powers(count, b)
    l = math.ceil(len(digits) / 2)
    return functools.reduce(lambda x, y: b * x + y, digits[:l]) + b ** (l - len(digits) % 2)


binary_palindrome_count = functools.partial(palindrome_count, b=2)
hex_palindrome_count = functools.partial(palindrome_count, b=16)


def cw_binary_count(comment_body):
    bits = parsing.extract_count_string(comment_body, base=2)
    shorter_counts = 2 ** len(bits) - 1
    fewer_ones = sum(int(binom(len(bits), ones)) for ones in range(bits.count("1")))
    earlier_counts = 0
    ones = 1
    for i, bit in enumerate(reversed(bits)):
        if bit == "1":
            earlier_counts += int(binom(i, ones))
            ones += 1
    return shorter_counts + fewer_ones + earlier_counts


@functools.cache
def mahonian(n, k):
    if n == 1 and k == 0:
        return 1
    if n < 0 or k < 0 or k > n * (n - 1) / 2:
        return 0
    return mahonian(n, k - 1) + mahonian(n - 1, k) - mahonian(n - 1, k - n)


# See https://old.reddit.com/r/counting/comments/18z0of2/free_talk_friday_436/kgii6c0/
def cw_factoradic_count(comment_body, base=10):
    count_string = parsing.extract_count_string(comment_body, base=base)
    digits = [int(char, base) for char in count_string]

    shorter_counts = sum(math.factorial(d + 1) for d in range(len(digits)))
    smaller_weights = sum(mahonian(len(digits) + 1, w) for w in range(sum(digits)))

    earlier_counts = 0
    k = 1
    for i, digit in enumerate(reversed(digits)):
        while digit > 0:
            earlier_counts += mahonian(i + 1, k)
            k += 1
            digit -= 1

    return shorter_counts + smaller_weights + earlier_counts


def update_dates(count, chain, previous=False):
    regex = r"([,\d]+)$"  # All digits at the end of the line, plus optional separators
    for submission in chain:
        year = int(re.search(regex, submission.title).group().replace(",", ""))
        if previous:
            year = year - 3
        length = 1095 + any(map(utils.is_leap_year, range(year, year + 3)))
        count += length
    return count


update_previous_dates = functools.partial(update_dates, previous=True)


def update_from_traversal(count, chain):
    for thread in chain[1:]:
        _, get_id = tn.find_previous_submission(thread)
        comments = tn.fetch_comments(get_id)
        count += len(comments)
    return count


known_threads = {
    "-illion": SideThread(form=illion_form, length=1000),
    "2d20 experimental v theoretical": SideThread(form=d20_form, length=1000),
    "balanced ternary": SideThread(form=balanced_ternary, length=729),
    "base 16 roman": SideThread(form=roman_numeral),
    "base 2i": SideThread(form=base_n(4), comment_to_count=gaussian_integer_count),
    "beenary": SideThread(length=1024, form=validate_from_tokens(["bee", "movie"])),
    "bijective base 2": SideThread(form=base_n(3), length=1024),
    "binary encoded decimal": SideThread(form=base_n(2), comment_to_count=bcd_count),
    "binary encoded hexadecimal": SideThread(form=base_n(2), length=1024),
    "binary palindromes": SideThread(form=base_n(2), comment_to_count=binary_palindrome_count),
    "by 3s in base 7": SideThread(form=base_n(7)),
    "by 3s": SideThread(comment_to_count=by_ns_count(3)),
    "by 4s": SideThread(comment_to_count=by_ns_count(4)),
    "by 5s": SideThread(comment_to_count=by_ns_count(5)),
    "by 7s": SideThread(comment_to_count=by_ns_count(7)),
    "by 99s": SideThread(comment_to_count=by_ns_count(99)),
    "collatz conjecture": SideThread(comment_to_count=collatz_count, form=base_10),
    "colored squares": SideThread(form=colored_squares_form, length=729),
    "constant sum factoradic": SideThread(form=base_10, comment_to_count=cw_factoradic_count),
    "constant weight binary": SideThread(form=base_n(2), comment_to_count=cw_binary_count),
    "cyclical bases": SideThread(form=base_n(16)),
    "dates": SideThread(form=base_10, update_function=update_dates),
    "decimal encoded sexagesimal": SideThread(length=900, form=base_10),
    "dollars and cents": SideThread(form=base_n(4)),
    "double increasing": SideThread(form=base_10, comment_to_count=increasing_type_count(2)),
    "fast or slow": SideThread(rule=FastOrSlow()),
    "four fours": SideThread(form=validate_from_tokens("4")),
    "hexadecimal palindromes": SideThread(form=base_n(16), comment_to_count=hex_palindrome_count),
    "increasing sequences": SideThread(form=base_10, comment_to_count=increasing_type_count(1)),
    "invisible numbers": SideThread(form=base_n(10, strip_links=False)),
    "isenary": SideThread(form=isenary_form, comment_to_count=isenary_count),
    "japanese": SideThread(form=validate_from_tokens("一二三四五六七八九十百千")),
    "letter permutations": SideThread(comment_to_count=letter_permutation_count),
    "mayan numerals": SideThread(length=800, form=mayan_form),
    "no repeating letters": SideThread(comment_to_count=nrl_count),
    "o/l binary": SideThread(form=validate_from_tokens("ol"), length=1024),
    "once per thread": SideThread(form=base_10, rule=CountingRule(wait_n=None)),
    "only double counting": SideThread(form=base_10, rule=OnlyDoubleCounting()),
    "ordered pairs": SideThread(form=base_10, comment_to_count=ordered_pairs_count),
    "palindromes": SideThread(form=base_10, comment_to_count=palindrome_count),
    "parentheses": SideThread(form=parentheses_form),
    "periodic table": SideThread(form=element_form, comment_to_count=element_count),
    "permutations": SideThread(form=base_10, comment_to_count=permutation_count),
    "previous dates": SideThread(form=base_10, update_function=update_previous_dates),
    "planetary octal": SideThread(comment_to_count=planetary_count, form=planetary_form),
    "powerball": SideThread(comment_to_count=powerball_count, form=base_10),
    "rainbow": SideThread(comment_to_count=rainbow_count, form=rainbow_form),
    "reddit usernames": SideThread(length=722, form=reddit_username_form),
    "rgb values": SideThread(form=base_10, comment_to_count=rgb_count),
    "roman progressbar": SideThread(form=roman_numeral),
    "roman": SideThread(form=roman_numeral),
    "slow": SideThread(form=base_10, rule=CountingRule(thread_time=MINUTE)),
    "slower": SideThread(form=base_10, rule=CountingRule(user_time=HOUR)),
    "slowestest": SideThread(form=base_10, rule=CountingRule(thread_time=HOUR, user_time=DAY)),
    "symbols": SideThread(form=validate_from_tokens("!@#$%^&*()")),
    "throwaways": SideThread(form=throwaway_form),
    "triple increasing": SideThread(form=base_10, comment_to_count=increasing_type_count(3)),
    "twitter handles": SideThread(length=1369, form=twitter_form),
    "unary": SideThread(form=validate_from_tokens("|")),
    "unicode": SideThread(form=base_n(16), length=1024),
    "us states": SideThread(form=us_states_form, comment_to_count=us_states_count),
    "using 12345": SideThread(form=validate_from_tokens("12345")),
    "valid brainfuck programs": SideThread(form=brainfuck),
    "wait 10": SideThread(form=base_10, rule=CountingRule(wait_n=10)),
    "wait 2 - letters": SideThread(rule=CountingRule(wait_n=2)),
    "wait 2": SideThread(form=base_10, rule=CountingRule(wait_n=2)),
    "wait 3": SideThread(form=base_10, rule=CountingRule(wait_n=3)),
    "wait 4": SideThread(form=base_10, rule=CountingRule(wait_n=4)),
    "wait 5s": SideThread(form=base_10, rule=CountingRule(thread_time=5)),
    "wait 9": SideThread(form=base_10, rule=CountingRule(wait_n=9)),
    "wave": SideThread(form=base_10, comment_to_count=wave_count),
}

known_threads.update(dfa_threads)

base_n_threads = {
    f"base {n}": SideThread(form=base_n(n), comment_to_count=base_n_count(n)) for n in range(2, 37)
}
known_threads.update(base_n_threads)

known_threads.update(
    {
        thread: SideThread(form=base_10, comment_to_count=base_n_count(10))
        for thread in ["by meters", "sheep", "word association"]
    }
)
# See: https://www.reddit.com/r/counting/comments/o7ko8r/free_talk_friday_304/h3c7433/?context=3


def by_x_count(comment_body: str, x=1.0):
    regex = r"([-+]?[\.,\s0-9]*)"
    body = parsing.normalize_comment_body(comment_body)
    count_string = re.search(regex, body).groups()[0]
    count_string = "".join(count_string.split()).replace(",", "")
    return int(float(count_string) / x)


by_xs = [0.01, 0.02, 0.025, 0.05, 6, 7, 8, 10, 11, 12, 20, 23, 29, 40, 50, 64, 69, 123, 1000]
known_threads.update(
    {
        f"by {x}s": SideThread(form=base_10, comment_to_count=functools.partial(by_x_count, x=x))
        for x in by_xs
    }
)

default_threads = [
    "10 at a time",
    "3 or fewer palindromes",
    "69, 420, or 666",
    "age",
    "all even or all odd",
    "by 2s even",
    "by 2s odd",
    "california license plates",
    "chess matches",
    "decimal",
    "four squares",
    "n read as base n number",
    "negative numbers",
    "powers of 2",
    "previous dates",
    "prime factorization",
    "prime numbers",
    "rational numbers",
    "rotational symmetry",
    "scientific notation",
    "street view counting",
    "thread completion",
    "top subreddits",
    "triangular numbers",
    "unordered consecutive digits",
    "william the conqueror",
]
known_threads.update(
    {thread_name: SideThread(form=base_10, length=1000) for thread_name in default_threads}
)

default_threads = {
    "eban": 800,
    "factoradic": 720,
    "feet and inches": 600,
    "hoi4 states": 806,
    "ipv4": 1024,
    "lucas numbers": 200,
    "seconds minutes hours": 1200,
    "time": 900,
}
known_threads.update(
    {key: SideThread(form=base_10, length=length) for key, length in default_threads.items()}
)


no_validation = {
    "acronyms": 676,
    "base 40": 1600,
    "base 60": 900,
    "base 62": 992,
    "base 64": 1024,
    "base 93": 930,
    "bijective base 205": 1025,
    "cards": 676,
    "degrees": 900,
    "iterate each letter": None,
    "letters": 676,
    "musical notes": 1008,
    "octal letter stack": 1024,
    "palindromes - letters": 676,
    "permutations - letters": None,
    "previous_dates": None,
    "qwerty alphabet": 676,
    "youtube": 1024,
}

known_threads.update({k: SideThread(length=v) for k, v in no_validation.items()})

default_thread_varying_length = [
    "2d tug of war",
    "boost 5",
    "by battery percentage",
    "by coad rank",
    "by comment karma",
    "by counters met irl",
    "by day of the week",
    "by day of the year",
    "by digits in total karma",
    "by gme increase/decrease",
    "by hoc rank",
    "by how well your day is going",
    "by length of username",
    "by number of post upvotes",
    "by random number (1-1000)",
    "by random number",
    "by timestamp seconds",
    "check-in streak",
    "nim",
    "pick from five",
    "post karma",
    "total karma",
    "tug of war",
]

default_thread_unknown_length = [
    "base of previous digit",
    "by list size",
    "by number of digits squared",
    "divisors",
]


def get_side_thread(thread_name):
    """Return the properties of the side thread with name thread_name"""
    if thread_name in known_threads:
        return known_threads[thread_name]
    if thread_name in default_thread_unknown_length:
        return SideThread(form=base_10)
    if thread_name in default_thread_varying_length:
        return SideThread(update_function=update_from_traversal, form=base_10)
    if thread_name != "default":
        printer.info(
            (
                "No rule found for %s. Not validating comment contents. "
                "Assuming n=1000 and no double counting."
            ),
            thread_name,
        )
    return SideThread()


config = configparser.ConfigParser()
config.read(os.path.join(module_dir, "side_threads.ini"))
known_thread_ids = config["threads"]



import logging
from collections import defaultdict, deque

from praw.exceptions import ClientException
from praw.models import MoreComments
from prawcore.exceptions import ServerError

from rcounting import parsing, utils

printer = logging.getLogger(__name__)


class Comment:
    """
    A lightweight class for combining a reddit comment object with a tree object
    """

    def __init__(self, comment, tree):
        self.comment = comment
        self.tree = tree
        self.created_utc = comment.created_utc
        self.removed = getattr(comment, "removed", False)
        self.body = comment.body
        self.id = comment.id
        self.link_id = comment.link_id
        self.author = str(comment.author)
        self.is_root = comment.is_root

    def walk_up_tree(self, *args, **kwargs):
        return self.tree.walk_up_tree(self, *args, **kwargs)

    @property
    def depth(self):
        return self.tree.find_depth(self)


class Tree:
    """
    A class for dealing with tree structures.

    The tree is represented as a dict, where y = tree[x] means y is the parent of x.

    Only the node ids are stored in this structure,the rest of the information
    about each node is stored in an auxiliary nodes dict
    """

    def __init__(self, nodes, tree):
        self.tree = tree
        self.nodes = nodes
        self.depths = {}

    def __contains__(self, node):
        node_id = extract_id(node)
        return node_id in self.tree.keys() or node_id in self.tree.values()

    def __len__(self):
        return len(self.tree.keys())

    @property
    def reversed_tree(self):
        return edges_to_tree([(parent, child) for child, parent in self.tree.items()])

    def parent(self, node):
        parent_id = self.tree[extract_id(node)]
        return self.node(parent_id)

    def find_children(self, node):
        return [self.node(x) for x in self.reversed_tree[extract_id(node)]]

    def walk_up_tree(self, node, limit=None, cutoff=None):
        """Navigate the tree from node to root"""
        if isinstance(node, str):
            try:
                node = self.node(node)
            except KeyError:
                return None
        if node.id not in self.tree and node.id not in self.nodes:
            return None
        nodes = [node]
        counter = 1
        while node.id in self.tree and not getattr(node, "is_root", False):
            if limit is not None and counter >= limit:
                break
            if cutoff is not None and getattr(node, "created_utc", float("inf")) < cutoff:
                nodes = nodes[:-1]
                break
            node = self.parent(node)
            nodes.append(node)
            counter += 1
        return nodes

    def walk_down_tree(self, node):
        """
        Navigate the tree from node to leaf, taking the earliest child each
        time there's a choice
        """
        if isinstance(node, str):
            node = self.node(node)
        if node.id not in self.nodes and node.id not in self.reversed_tree:
            return [node]
        result = [node]
        while node.id in self.reversed_tree:
            node = self.find_children(node)[0]
            result.append(node)
        return result

    def node(self, node_id):
        return self.nodes[node_id]

    def delete_edge(self, child, parent):
        child_id = extract_id(child)
        parent_id = extract_id(parent)
        if self.tree.get(child_id, float("nan")) == parent_id:
            del self.tree[child_id]

    def delete_node(self, node):
        """Given a node id, this will delete the node from the tree if it is
        present. That means deleting it from the list of known nodes and (if
        the node is not a root node) deleting it from the child -> parent
        dictionary. If the node is not present, an exception will be raised."""

        node_id = extract_id(node)
        if node_id in self.nodes:
            del self.nodes[node_id]
        if node_id in self.tree:
            del self.tree[node_id]

    def delete_subtree(self, node):
        """Delete the entire subtree rooted at the `node`"""
        queue = deque([node])
        while queue:
            node = queue.popleft()
            queue.extend(self.find_children(node))
            self.delete_node(node)

    def find_depth(self, node):
        """
        Find the depth of a node.

        The root nodes have depth 0. Otherwise, each node is one deeper than its parent.
        """
        node_id = extract_id(node)
        if node_id in self.root_ids:
            return 1
        if node_id in self.depths:
            return self.depths[node_id]
        depth = 1 + self.find_depth(self.parent(node_id))
        self.depths[node_id] = depth
        return depth

    @property
    def deepest_node(self):
        max_depth = 0
        result = None
        for leaf in self.leaves:
            depth = self.find_depth(leaf)
            if depth > max_depth:
                max_depth = depth
                result = leaf
        return result

    @property
    def leaves(self):
        leaf_ids = set(self.nodes.keys()) - set(self.tree.values())
        return [self.node(leaf_id) for leaf_id in leaf_ids]

    @property
    def roots(self):
        return [self.node(root_id) for root_id in self.root_ids]

    @property
    def root_ids(self):
        root_ids = (set(self.nodes.keys()) | set(self.tree.values())) - set(self.tree.keys())
        root_ids = [[x] if x in self.nodes else self.reversed_tree[x] for x in root_ids]
        return [root_id for ids in root_ids for root_id in ids]

    def add_nodes(self, new_nodes, new_tree):
        self.tree.update(new_tree)
        self.nodes.update(new_nodes)


class CommentTree(Tree):
    """
    A class representing the comment tree

    In addition to all the things the superclass can do, this one can use
    a reddit instance to get information about missing comments. That means
    that many of the methods for finding parents & children have to be overridden.
    """

    def __init__(self, comments=None, reddit=None, get_missing_replies=True):
        if comments is None:
            comments = []
        tree = {x.id: x.parent_id[3:] for x in comments if not x.is_root}
        comments = {x.id: x for x in comments}
        super().__init__(comments, tree)
        self.reddit = reddit
        self.get_missing_replies = get_missing_replies
        # logger levels are 10, 20, 30, where 10 is most verbose
        self.refresh_counter = [0, 5, 2][3 - int(printer.getEffectiveLevel() // 10)]
        self._parent_counter, self._child_counter = 0, 0
        self.comment = self.node

    def node(self, node_id):
        if node_id not in self.tree:
            self.add_missing_parents(node_id)
        return Comment(super().node(node_id), self)

    def add_comments(self, comments):
        new_comments = {x.id: x for x in comments}
        new_tree = {x.id: x.parent_id[3:] for x in comments if not x.is_root}
        super().add_nodes(new_comments, new_tree)

    @property
    def comments(self):
        return self.nodes.values()

    def add_missing_parents(self, comment_id):
        comments = []
        if self.reddit is None:
            return
        praw_comment = self.reddit.comment(comment_id)
        if praw_comment.is_root:
            self.add_comments([praw_comment])
            return
        try:
            praw_comment.refresh()
            if self._parent_counter == 0:
                printer.info("Fetching ancestors of comment %s", normalise(praw_comment.body))
                self._parent_counter = self.refresh_counter
            else:
                self._parent_counter -= 1
        except (ClientException, ServerError) as e:
            printer.warning("Unable to refresh %s", comment_id)
            print(e)
        for _ in range(9):
            comments.append(praw_comment)
            if praw_comment.is_root:
                break
            praw_comment = praw_comment.parent()
        self.add_comments(comments)

    def fill_gaps(self):
        for node in self.roots:
            if not node.is_root:
                node.walk_up_tree()

    def find_children(self, node):
        """
        Return a sorted list of comments that are the direct children of
        `node`. The list is sorted so that all deleted comments appear after
        all non-deleted comments and then by timestamp, so that earlier
        comments appear ahead of later ones.

        The goal is to play nice with a depth-first expansion of the comment tree.

        """
        node_id = extract_id(node)
        children = [self.comment(x) for x in self.reversed_tree[node_id]]
        if not children and self.get_missing_replies:
            children = self.add_missing_replies(node_id)

        def comment_order(comment):
            return (comment.body in utils.deleted_phrases, comment.created_utc)

        return sorted(children, key=comment_order)

    def add_missing_replies(self, comment):
        comment_id = extract_id(comment)
        if self.reddit is None:
            return []
        praw_comment = self.reddit.comment(comment_id)
        if comment_id not in self.nodes:
            self.add_comments([praw_comment])

        praw_comment.refresh()
        replies = find_all_replies(praw_comment)
        if replies:
            self.add_comments(replies)
            return [self.comment(x.id) for x in replies]
        return []

    def prune(self, side_thread, comment=None):
        """
        Use a side thread object to remove invalid comments and their descendants
        from the comment tree.
        """
        if comment is None:
            nodes = self.roots
        else:
            nodes = self.find_children(comment)
        queue = deque([(node, side_thread.get_history(node)) for node in nodes])
        while queue:
            node, history = queue.popleft()
            if node.removed:
                self.delete_subtree(node)
                continue
            is_valid, new_history = side_thread.is_valid_count(node, history)
            if is_valid:
                queue.extend([(x, new_history) for x in self.find_children(node)])
            else:
                self.delete_subtree(node)


class SubmissionTree(Tree):
    """
    A tree that tracks submissions.

    It currently can only keep track of whether or not a submission is archived.
    """

    def __init__(self, submissions, submission_tree, reddit=None):
        self.reddit = reddit
        super().__init__(submissions, submission_tree)

    def is_archived(self, submission):
        return extract_id(submission) not in self.nodes

    def node(self, node_id):
        try:
            return super().node(node_id)
        except KeyError:
            if self.reddit is not None:
                return self.reddit.submission(node_id)
            raise


def edges_to_tree(edges):
    """
    Popupate a tree dictionary from a list of edges
    """
    tree = defaultdict(list)
    for source, dest in edges:
        tree[source].append(dest)
    return tree


def comment_to_dict(comment):
    return {
        "username": str(comment.author),
        "timestamp": comment.created_utc,
        "comment_id": comment.id,
        "submission_id": comment.link_id[3:],
        "body": comment.body,
    }


def submission_to_dict(submission):
    return {
        "username": str(submission.author),
        "timestamp": submission.created_utc,
        "submission_id": submission.id,
        "body": submission.selftext,
        "title": submission.title,
    }


def normalise(body):
    return parsing.strip_markdown_links(body.split("\n")[0])


def extract_id(node):
    if hasattr(node, "id"):
        return node.id
    return node


def find_all_replies(comment):
    """Return a list of all replies to a comment, with each `MoreComments`
    instance expanded. The comments might be out of order, and due to reddit
    weirdness a comment might appear multiple times. That's OK, since the tree
    class can handle reconstructing the order and doesn't care about repeats."""

    comments = comment.replies.list()
    replies = []
    submission = comment.submission
    while comments:
        item = comments.pop(0)
        if isinstance(item, MoreComments):
            if item.submission is None:
                item.submission = submission
            comments = item.comments(update=False).list() + comments
        else:
            replies.append(item)
    return replies



import bisect
import copy
import datetime
import itertools
import logging
import time

from rcounting import models, parsing
from rcounting import side_threads as st
from rcounting import utils

printer = logging.getLogger(__name__)


def load_wiki_page(subreddit, location, kind="directory"):
    """
    Load the wiki page at reddit.com/r/subreddit/wiki/location

    Normalise the newlines, and parse it into a list of paragraphs.
    """
    wiki_page = subreddit.wiki[location]
    document = wiki_page.content_md.replace("\r\n", "\n")
    return Directory(parsing.parse_directory_page(document), kind)


def title_from_first_comment(submission):
    """
    Return the body of the first comment of the submission, appropriately normalised:

    - Markdown links are stripped.
    - Only the first line is considered.
    """
    comment = sorted(list(submission.comments), key=lambda x: x.created_utc)[0]
    body = comment.body.split("\n")[0]
    return parsing.normalise_title(parsing.strip_markdown_links(body))


class Row:
    """
    A class corresponding to a row in a markdown table in the directory.

    A row has the following information associated with it:
      - The human-readable name of the thread. E.g. 'No repeating digits'
      - The submission id of the first submission, used to identify the side thread
      - The title of the current submission in the thread
      - The id of the current submission in the thread
      - The id of the latest comment in the current submission
      - The total number of counts made in the thread

    The core of this class is the `update` method, which updates
    the last four pieces of information listed.
    """

    def __init__(self, name, first_submission, title, submission_id, comment_id, count):
        """
        Initialise a new row with all the necessary information. Associate a side
        thread object with the row.
        """
        self.archived = False
        self.name = name
        self.first_submission = first_submission
        self.title = parsing.normalise_title(title)
        self.initial_submission_id = submission_id
        self.initial_comment_id = comment_id
        self.count_string = count
        try:
            self.count = parsing.find_count_in_text(self.count_string.replace("-", "0"))
        except ValueError:
            self.count = None
        self.is_approximate = self.count_string[0] == "~"
        self.starred_count = self.count_string[-1] == "*"
        self.thread_type = st.known_thread_ids.get(self.first_submission, fallback="default")
        self.submission = None
        self.comment = None

    def __str__(self):
        return (
            f"[{self.name}](/comments/{self.first_submission}) | "
            f"[{self.title}]({self.link}) | {self.count_string}"
        )

    def order_tuple(self):
        """
        The order in which total counts should be compared.

        Tuple comparison works in the dictionary order, so a < b means that one
        of the following is true
          - a[0] < b[0]
          - a[0] == b[0] and a[1:] < b[1:]
        """
        return (self.count, self.starred_count, self.is_approximate)

    def __lt__(self, other):
        return self.order_tuple() < other.order_tuple()

    @property
    def submission_id(self):
        return self.submission.id if self.submission is not None else self.initial_submission_id

    @property
    def comment_id(self):
        return self.comment.id if self.comment is not None else self.initial_comment_id

    @property
    def link(self):
        """Set the full link if we have it, otherwise just link the submission."""
        if self.comment_id is not None:
            return f"/comments/{self.submission_id}/_/{self.comment_id}?context=3"
        return f"/comments/{self.submission_id}"

    def update_title(self):
        """
        Set the title of the current submission.

        If it's the first submission, set the title from the first comment.
        Otherwise, treat the submission title as a | delimited list and use all
        but the first section.

        Then normalise the title before setting it
        """
        if self.first_submission == self.submission.id:
            self.title = title_from_first_comment(self.submission)
            return
        sections = self.submission.title.split("|")
        if len(sections) > 1:
            title = "|".join(sections[1:]).strip()
        else:
            title = title_from_first_comment(self.submission)
        self.title = parsing.normalise_title(title)

    def update_count(self, chain, side_thread):
        """
        Use the side thread get an updated tally of how many counts have been
        made in the thread."""
        try:
            count = side_thread.update_count(self.count, chain)
        except (ValueError, IndexError):
            count = None
        self.count_string = self.format_count(count)
        if count is not None:
            self.count = count
        else:
            self.starred_count = True

    def format_count(self, count):
        """Add asterisks and tildes to erroneous and approximate counts"""
        if count is None:
            return self.count_string + "*"
        if count == 0:
            return "-"
        if self.is_approximate:
            return f"~{count:,d}"
        return f"{count:,d}"

    def update(self, submission_tree, deepest_comment=False):
        """Find the latest comment in the latest submission of the side thread
        represented by this row.

        Parameters:

        submission_tree: A models.Tree object representing which
        submissions are linked to which. If no mistakes have been made, this
        should just be a series of straight line chains

        deepest_comment: A flag used to say that the function should find the
        deepest comment overall, rather than the leaf of the earliest valid chain.
        Earliest is defined according to the order

        a < b if a is an ancestor of b
        a < b if a and b have the same parent and a was posted before b
        a < b if π(a) < π(b), where π(a) is the oldest ancestor of a which is
        not an ancestor of b, and similarly for b.

        This is used for new threads, where non-count comments are frequently
        posted either as early top-level comments, or as replies to a top level
        comment.

        """
        side_thread = st.get_side_thread(self.thread_type)
        printer.debug("Updating side thread: %s", self.thread_type)
        if self.thread_type == "default":
            printer.warning(
                "No rule found for %s. Not validating comment contents. "
                "Assuming n=1000 and no double counting.",
                self.name,
            )

        chain = submission_tree.walk_down_tree(submission_tree.node(self.submission_id))
        self.submission = chain[-1]

        if len(chain) > 1:
            self.initial_comment_id = None

        comments = models.CommentTree(reddit=submission_tree.reddit, get_missing_replies=False)
        if deepest_comment:
            for comment in self.submission.comments:
                comments.add_missing_replies(comment)
        elif self.comment_id is None:
            comment = next(filter(side_thread.looks_like_count, self.submission.comments))
            comments.add_missing_replies(comment)
        else:
            comments.add_missing_replies(self.comment_id)
            comment = comments.node(self.comment_id)
        comments.prune(side_thread, self.comment_id)
        if deepest_comment:
            comment = comments.deepest_node.walk_up_tree(limit=3)[-1]
        else:
            comment_chain = comments.walk_down_tree(comment)
            comment = comment_chain[-3 if len(comment_chain) >= 3 else 0]

        self.comment = comment
        if len(chain) > 1:
            # If there's really a new thread, the title & count need updating
            self.update_count(chain, side_thread)
            self.update_title()
        if submission_tree.is_archived(self.submission):
            comment = comment.walk_up_tree(limit=5)[-1]
            dt = datetime.timedelta(days=60)
            now = datetime.datetime.now(datetime.timezone.utc)
            if (
                now
                - datetime.datetime.fromtimestamp(comment.created_utc, tz=datetime.timezone.utc)
                > dt
            ):
                self.archived = True


class Paragraph:
    """A class to hold either a text paragraph, or a markdown table"""

    def __init__(self, tagged_text, kind="directory"):
        self.tag, self.contents = tagged_text
        if self.tag == "table":
            if not self.contents:
                self.contents = []
            self.contents = [Row(*x) if hasattr(x, "__iter__") else x for x in self.contents]
        self.kind = kind

    def sort(self, *args, **kwargs):
        if self.tag != "text":
            self.contents.sort(*args, **kwargs)

    def __str__(self):
        if self.tag == "text":
            return self.contents
        labels = {"directory": "Current", "archive": "Last"}
        header = [
            "⠀" * 6 + "Name &amp; Initial Thread" + "⠀" * 6,
            "⠀" * 6 + f"{labels[self.kind]} Thread" + "⠀" * 6,
            "⠀" * 1 + "# of Counts" + "⠀" * 1,
        ]
        header = [" | ".join(header), ":--:|:--:|--:"]
        rows = self.contents
        if self.kind == "directory":
            rows = [x for x in self.contents if not x.archived]
        return "\n".join(header + [str(x) for x in rows])

    def is_misc_table_heading(self):
        if not isinstance(self.contents, str):
            return False
        return "new" in self.contents.lower() and "revived" in self.contents.lower()

    def delete_row_if_present(self, row_to_delete):
        if self.tag == "text":
            return
        self.contents = [row for row in self.contents if id(row) != id(row_to_delete)]


class Directory:
    """A class to hold the state associated with the thread directory.

    At its core, the directory is a collection of paragraphs, each of which is
    either a table with information about side threads, or a text paragraph.

    Updating the directory means updating each paragraph, then checking for new
    submissions, and finally checking for revived submissions. There's a fair
    bit of fiddling to keep track of, which is why the whole thing is
    encapsulated in this class.

    """

    def __init__(self, paragraphs, kind="directory", archive=None):
        self.paragraphs = [Paragraph(x, kind) for x in paragraphs]
        self.known_submissions = {x.submission_id for x in self.rows}
        if archive is None:
            archive = {}
        self.archive = archive
        self.updated_archive = False
        self.header = paragraphs[0][1]

    @property
    def tables(self):
        return [x for x in self.paragraphs if x.tag == "table"]

    @property
    def rows(self):
        return utils.flatten([x.contents for x in self.tables])

    @property
    def first_submissions(self):
        return [x.first_submission for x in self.rows]

    def __str__(self):
        return "\n\n".join(str(x) for x in self.paragraphs)

    def set_archive(self, archive):
        archive = {x.submission_id: x for x in archive.rows}
        self.archive = archive

    def update(self, tree, new_submission_ids, sleep=0):
        printer.info("Updating tables")
        self.update_existing_rows(tree, sleep)
        self.add_last_table()
        printer.info("Updating new threads")
        new_submissions = self.find_new_submissions(tree, new_submission_ids)
        printer.info("Finding revived threads")
        revived_submissions = self.find_revived_submissions(tree, new_submission_ids)
        new_top_25, promoted, demoted = self.find_new_top_25()
        self.tables[1].contents = new_top_25
        for row in promoted:
            for table in self.tables[2:]:
                table.delete_row_if_present(row)
        self.paragraphs[-2].contents += new_submissions + revived_submissions + demoted
        self.paragraphs[-2].sort(key=lambda x: parsing.name_sort(x.name))
        archived_rows = [row for row in self.rows if row.archived]
        if archived_rows:
            self.updated_archive = True
            self.archive.update({row.submission_id: row for row in archived_rows})

    def update_existing_rows(self, tree, sleep=0):
        """Update every row in the main directory page"""
        for row in self.rows:
            try:
                row.update(tree)
                if sleep:
                    time.sleep(sleep)
            except Exception:  # pylint: disable=broad-except
                printer.warning("Unable to update thread %s", row.title)
                raise
        self.known_submissions |= {row.submission_id for row in self.rows}

    def add_last_table(self):
        if not self.paragraphs[-3].is_misc_table_heading():
            self.paragraphs = (
                self.paragraphs[:-1]
                + [Paragraph(["text", "\n## New and Revived Threads"]), Paragraph(["table", []])]
                + self.paragraphs[-1:]
            )

    def find_new_submissions(self, tree, new_submission_ids):
        """
        Make a list of update rows corresponding to new submissions

        If a row cannot be updated or, don't include it.
        The same goes for new submissions with only a few comments on them.
        """
        result = []
        for submission_id in new_submission_ids - self.known_submissions:
            first_submission = tree.walk_up_tree(submission_id)[-1]
            name = f'**{first_submission.title.split("|")[0].strip()}**'
            try:
                title = title_from_first_comment(first_submission)
            except IndexError:
                continue
            row = Row(name, first_submission.id, title, first_submission.id, None, "-")
            try:
                row.update(tree, deepest_comment=True)
            except Exception:  # pylint: disable=broad-except
                printer.warning("Unable to update new thread %s", row.title)
                raise
            n_authors = len(set(x.author for x in row.comment.walk_up_tree()))
            is_long_chain = row.comment.depth >= 50 and n_authors >= 5
            if is_long_chain or row.submission_id != first_submission.id:
                result.append(row)
        return result

    def find_revived_submissions(self, tree, new_submission_ids):
        """
        Make a list of updated rows corresponding to revived threads.

        If a row cannot be updated or, don't include it.
        The same goes for new submissions with only a few comments on them.
        """
        revivals = []
        revived_threads = {x.id for x in tree.leaves} - new_submission_ids - self.known_submissions
        for thread in revived_threads:
            chain = tree.walk_up_tree(thread)
            for submission in chain:
                if submission.id in self.archive:
                    row = copy.copy(self.archive[submission.id])
                    try:
                        row.update(tree, deepest_comment=True)
                    except Exception:  # pylint: disable=broad-except
                        printer.warning("Unable to update revived thread %s", row.title)
                        raise
                    if row.comment.depth >= 20 or len(chain) > 2:
                        revivals.append(row)
                        del self.archive[submission.id]
                        self.updated_archive = True
                    break
        return revivals

    def archive2string(self):
        archived_rows = list(self.archive.values())
        printer.info("Updating archive at /r/counting/wiki/directory/archive")
        archived_rows.sort(key=lambda x: parsing.name_sort(x.name))
        splits = ["A", "D", "I", "P", "T", "["]
        titles = [f"\n### {splits[idx]}-{chr(ord(x) - 1)}" for idx, x in enumerate(splits[1:])]
        keys = [parsing.name_sort(x.name) for x in archived_rows]
        indices = [bisect.bisect_left(keys, (split.lower(),)) for split in splits[1:-1]]
        parts = [
            str(Paragraph(["table", x], kind="archive"))
            for x in utils.partition(archived_rows, indices)
        ]
        archive = list(itertools.chain.from_iterable(zip(titles, parts)))
        return "\n\n".join(archive[1:])

    def top_25(self):
        rows = [row for row in sorted(self.rows[1:], reverse=True) if not row.archived]
        return rows[:25]

    def find_new_top_25(self):
        previous_top_25 = self.tables[1].contents
        previous_ids = {id(x) for x in previous_top_25}
        new_top_25 = self.top_25()
        new_ids = {id(x) for x in new_top_25}
        id_mapping = {id(x): x for x in previous_top_25 + new_top_25}
        promoted_ids = new_ids - previous_ids
        demoted_ids = previous_ids - new_ids
        return (
            new_top_25,
            [id_mapping[x] for x in promoted_ids],
            [id_mapping[x] for x in demoted_ids],
        )


def comment_to_row(comment) -> Row:
    """Takes a comment on a new side thread and returns a Row object that
    represents that comment under the assumption that:

    - The comment is the leaf comment in the submission of interest
    - The submission is the first one in a new side thread

    This is useful for when submissions need to be manually added to the directory.
    """
    submission = comment.submission
    name = f'**{submission.title.split("|")[0].strip()}**'
    title = title_from_first_comment(submission)
    return Row(name, submission.id, title, submission.id, comment.id, "-")



import pandas as pd

from rcounting.models import comment_to_dict
from rcounting.units import HOUR, MINUTE


class CountingRule:
    """
    A rules class. It knows how to do two things:

    - Get enough history to see whether a given comment is valid
    - Determine whether all counts in a history of counts are valid.

    Examples of things it's intended to validate are:
      - That users waited enough time since their own last comment before commenting again
      - That users waited enough time since the global last comment
      - That users let enough other counters go before them
    """

    def __init__(self, wait_n: int | None = 1, thread_time=0, user_time=0):
        self.n = wait_n
        self.thread_time = thread_time
        self.user_time = user_time

    def _valid_skip(self, history):
        n = self.n if self.n is not None else len(history)
        history = history.reset_index()
        skips = history.groupby("username")["index"].diff()
        return skips.isna() | (skips > n)

    def _valid_thread_time(self, history):
        if not self.thread_time:
            return True
        elapsed_time = history["timestamp"].diff()
        valid_time = elapsed_time.isna() | (elapsed_time >= self.thread_time)
        return valid_time

    def _valid_user_time(self, history):
        if not self.user_time:
            return True
        elapsed_user_time = history.groupby("username")["timestamp"].diff()
        valid_user_time = elapsed_user_time.isna() | (elapsed_user_time >= self.user_time)
        return valid_user_time

    def is_valid(self, history):
        return (
            self._valid_skip(history)
            & self._valid_thread_time(history)
            & self._valid_user_time(history)
        )

    def get_history(self, comment):
        limit = self.n + 1 if self.n is not None else self.n
        comments = comment.walk_up_tree(limit=limit)
        max_time = max(self.thread_time, self.user_time)
        while (
            not comments[-1].is_root
            and (comment.created_utc - comments[-1].created_utc) < max_time
        ):
            comments = comments[:-1] + comments[-1].walk_up_tree(limit=9)
        return pd.DataFrame([comment_to_dict(x) for x in comments[:0:-1]])


class FastOrSlow(CountingRule):
    """
    An special case of the rules class to account for a thread where the rule
    is not of the form 'wait at least X' before counting, but rather
    'wait at most X or at least Y'.

    """

    def __init__(self):
        super().__init__()

    def _valid_thread_time(self, history):
        elapsed_time = history["timestamp"].diff()
        valid_time = elapsed_time.isna() | (elapsed_time < 5 * MINUTE) | (elapsed_time >= HOUR)
        return valid_time


class OnlyDoubleCounting(CountingRule):
    """
    Only double counting is sufficiently strange that it gets its own class.

    A thread is valid if every user in the chain counts exactly twice in a row.
    """

    def is_valid(self, history):
        history = history.set_index("comment_id")
        history["mask"] = True
        unshifted = history.username.iloc[::2]
        up_shift = history.username.shift(-1).iloc[::2]
        up_mask = up_shift.isna() | (up_shift == unshifted)
        down_shift = history.username.shift().iloc[::2]
        down_mask = down_shift.isna() | (down_shift == unshifted)
        mask = up_mask if (up_mask.sum() > down_mask.sum()) else down_mask
        history.loc[mask.index, "mask"] = mask
        history.reset_index(inplace=True)
        return history["mask"]

    def get_history(self, comment):
        comments = comment.walk_up_tree(limit=2)[:0:-1]
        return pd.DataFrame([comment_to_dict(x) for x in comments])


default_rule = CountingRule()



# pylint: disable=import-outside-toplevel
"""A script to update the thread directory"""
import datetime
import logging

import click

from rcounting import configure_logging

printer = logging.getLogger("rcounting")

# Sometimes people include spurious links in posts. Here's a list of
# (link_from, link_to) which should be ignored.
spurious_edges = [
    ("r2h98h", "nyg22w"),
    ("11va8n8", "11siye8"),
    ("12p80i5", "11nouml"),
    ("1d1pyvo", "1crh3wj"),
    ("1e37w8n", "1diqgjm"),
]

# And some posts just shouldn't be tracked - here's a hardcoded list of those
spurious_nodes = ("12x4vpl", "13s4tu6")


@click.command()
@click.option(
    "--dry-run", is_flag=True, help="Write results to files instead of updating the wiki pages"
)
@click.option("-v", "--verbose", count=True, help="Print more output")
@click.option("-q", "--quiet", is_flag=True)
@click.option("--sleep", default=0)
def update_directory(quiet, verbose, dry_run, sleep):
    """
    Update the thread directory located at reddit.com/r/counting/wiki/directory.
    """

    from rcounting import thread_directory as td
    from rcounting import thread_navigation as tn
    from rcounting.reddit_interface import subreddit

    configure_logging.setup(printer, verbose, quiet)
    start = datetime.datetime.now()
    printer.info("Getting history")
    tree, new_submissions = tn.fetch_counting_history(subreddit, datetime.timedelta(days=187))
    for edge in spurious_edges:
        tree.delete_edge(*edge)

    for node in spurious_nodes:
        if node in tree:
            tree.delete_node(node)
    new_submissions = [x for x in new_submissions if x.id not in spurious_nodes]
    new_submission_ids = {tree.walk_down_tree(submission)[-1].id for submission in new_submissions}

    directory = td.load_wiki_page(subreddit, "directory")
    archive = td.load_wiki_page(subreddit, "directory/archive", kind="archive")
    directory.set_archive(archive)
    directory.update(tree, new_submission_ids, sleep)

    if not dry_run:
        subreddit.wiki["directory"].edit(content=str(directory), reason="Ran the update script")
    else:
        with open("directory.md", "w", encoding="utf8") as f:
            print(directory, file=f)

    if directory.updated_archive:
        archive = "\n\n".join([archive.header, directory.archive2string()])
        if not dry_run:
            subreddit.wiki["directory/archive"].edit(
                content=archive, reason="Ran the update script"
            )
        else:
            with open("archive.md", "w", encoding="utf8") as f:
                print(archive, file=f)
    end = datetime.datetime.now()
    printer.info("Running the script took %s", end - start)


if __name__ == "__main__":
    update_directory()  # pylint: disable=no-value-for-parameter



# pylint: disable=import-outside-toplevel
import datetime as dt

import click

from rcounting.counters import apply_alias
from rcounting.ftf import get_ftf_timestamp, is_within_threshold
from rcounting.parsing import parse_markdown_links


def find_manual_ftf(previous_ftf_poster, subreddit):
    submissions = []
    for submission in subreddit.new(limit=1000):
        if is_within_threshold(submission):
            submissions.append(submission)
        else:
            break
    candidate_ftfs = [
        submission for submission in submissions if "Free Talk Friday" in submission.title
    ]
    if not candidate_ftfs:
        return False
    if len(candidate_ftfs) > 1:
        for candidate_ftf in candidate_ftfs[::-1]:
            if candidate_ftf.author != previous_ftf_poster:
                return candidate_ftf
    return candidate_ftfs[-1]


def pprint(date):
    return date.strftime("%A %B %d, %Y")


def generate_new_title(previous_title):
    n = int(previous_title.split("#")[1])
    return f"Free Talk Friday #{n+1}"


def generate_new_body(previous_ftf_id, threshold_date, bot=True):

    ftf_body = (
        f"Continued from last week's FTF [here](/comments/{previous_ftf_id}/)\n\n"
        "It's that time of the week again. Speak anything on your mind! "
        "This thread is for talking about anything off-topic, "
        "be it your lives, your strava, your plans, your hobbies, studies, stats, "
        "pets, bears, hikes, dragons, trousers, travels, transit, cycling, family, "
        "colours, or anything you like or dislike, except politics\n\n"
        "Feel free to check out our [tidbits](https://redd.it/1bgwyht) thread "
        "and introduce yourself if you haven't already."
    )

    if not bot:
        return ftf_body
    bot_body = (
        "\n\n---\n\n"
        "*This post was made by a bot, because no-one else made a Free Talk Friday post "
        " before {} UTC 13:00. Anyone can post the FTF, so if you want to have your "
        "post pinned here for a week, just make one {} between UTC 07:00 and 13:00. "
        "The rules for these posts can be found in the "
        "[faq](/r/counting/wiki/faq/#wiki_3.6_free_talk_friday_posts). You can also check out "
        "our [directory](/r/counting/wiki/ftf_directory) of older posts for inspiration.*\n\n"
        "*If you have any questions or comments about the bot, feel free to write them below,"
        " or message the mods.*"
    )
    return ftf_body + bot_body.format(
        pprint(threshold_date), pprint(threshold_date + dt.timedelta(weeks=1))
    )


def make_directory_row(post):
    date = dt.date.fromtimestamp(post.created_utc)
    link = f"[FTF #{post.title.split('#')[1]}](/{post.id})"
    formatted_date = date.strftime("%b %d, %Y")
    author = apply_alias(str(post.author))
    return f"|{link}|{formatted_date}|{author}"


def update_directory(post, subreddit):
    row = make_directory_row(post)
    wiki = subreddit.wiki["ftf_directory"]
    contents_list = wiki.content_md.split("\n")
    links = [parse_markdown_links(x) for x in contents_list]
    known_posts = [x[0][1][1:] for x in links if x]
    if post.id not in known_posts:
        new_contents = "\n".join(contents_list + [row])
        wiki.edit(content=new_contents, reason="Added latest FTF")


@click.command(name="ftf")
@click.argument("subreddit", default="counting")
@click.option("--bot/--no-bot", default=True)
def pin_or_create_ftf(subreddit, bot):
    """
    Pin the earliest valid Free Talk Friday thread for this week in [subreddit].
    The subreddit is r/counting by default, but passing a different value lets you
    test things on a subreddit you control.

    If no FTF has been posted, create one, and then pin it.

    Also update the FTF directory with the newest FTF.
    """
    from rcounting.reddit_interface import reddit

    subreddit = reddit.subreddit(subreddit)

    previous_ftf_post = subreddit.sticky(number=2)
    threshold_timestamp = get_ftf_timestamp()

    if is_within_threshold(previous_ftf_post):
        ftf_post = previous_ftf_post
    else:
        ftf_post = find_manual_ftf(previous_ftf_post.author, subreddit)
        if not ftf_post:
            title = generate_new_title(previous_ftf_post.title)
            body = generate_new_body(previous_ftf_post.id, threshold_timestamp, bot)
            ftf_post = subreddit.submit(title=title, selftext=body)

        previous_ftf_post.mod.sticky(state=False)
        ftf_post.mod.approve()
        ftf_post.mod.sticky()
    ftf_post.mod.suggested_sort(sort="new")
    update_directory(ftf_post, subreddit)


if __name__ == "__main__":
    pin_or_create_ftf()  # pylint: disable=no-value-for-parameter



import functools
from typing import Callable, Iterable, Mapping

from fuzzywuzzy import fuzz

from rcounting import parsing


def by_ns_count(n):
    def comment_to_count(comment):
        count = parsing.find_count_in_text(comment)
        return int(count // n)

    return comment_to_count


def base_n_count(n: int):
    def comment_to_count(comment_body: str):
        return parsing.find_count_in_text(comment_body, base=n)

    return comment_to_count


def fuzzy_tokenize(comment_body, tokens, ignored_chars=">", threshold=80):
    line = comment_body.split("\n")[0]
    line = "".join(char for char in line if char not in ignored_chars)
    words = line.lower().strip().split()
    prefixes = list(set(y[0] for token in tokens if len(y := token.split()) > 1))
    complete_tokens = [token for token in tokens if len(token.split()) == 1]
    i = 0
    values = []
    while i < len(words):
        candidate = max(
            (fuzz.ratio(token, words[i]), token) for token in prefixes + complete_tokens
        )
        if candidate[1] in prefixes:
            if i + 1 == len(words):
                break
            candidate = max(
                (fuzz.ratio(token, " ".join(words[i : i + 2])), token) for token in tokens
            )
            i += 1
        if candidate[0] < threshold:
            break
        values.append(candidate[1])
        i += 1
    return values


def count_from_word_list(
    comment_body: str,
    alphabet: str | Iterable[str] | Mapping[str, int] = "0123456789",
    tokenize: Callable[[str, list[str]], list[str]] = fuzzy_tokenize,
    bijective=False,
    **kwargs,
) -> int:
    if not isinstance(alphabet, dict):
        alphabet = {k: p + int(bijective) for p, k in enumerate(alphabet)}
    base = len(set(alphabet.values()))
    tokens = tokenize(comment_body, list(alphabet.keys()), **kwargs)
    values = [alphabet[token] for token in tokens]
    return functools.reduce(lambda x, y: base * x + y, values, 0)



# pylint: disable=import-outside-toplevel,too-many-arguments,too-many-locals
"""Script for logging reddit submissions to either a database or a csv file"""
import logging
import time
from datetime import datetime
from pathlib import Path

import click
import pandas as pd
from prawcore.exceptions import TooManyRequests

printer = logging.getLogger("rcounting")


@click.command()
@click.argument("leaf_comment_id", default="")
@click.option("-n", "--n-threads", default=1, help="The number of submissions to log.")
@click.option("--all", "-a", "all_counts", is_flag=True, help="Log all threads. Can take a while!")
@click.option(
    "--filename",
    "-f",
    type=click.Path(path_type=Path),
    help=(
        "What file to write output to. If none is specified, counting.sqlite is used as default in"
        " sql mode, and the base count is used in csv mode."
    ),
)
@click.option(
    "-o",
    "--output-directory",
    default=".",
    type=click.Path(path_type=Path),
    help="The directory to use for output. Default is the current working directory",
)
@click.option(
    "--sql/--csv",
    default=False,
    help="Write submissions to csv files (one per thread) or to a database.",
)
@click.option(
    "--side-thread/--main",
    "-s/-m",
    default=False,
    help=(
        "Log the main thread or a side thread. Get validation is "
        "switched off for side threads, and only sqlite output is supported"
    ),
)
@click.option("--verbose", "-v", count=True, help="Print more output")
@click.option("--quiet", "-q", is_flag=True, default=False, help="Suppress output")
def log(
    leaf_comment_id,
    all_counts,
    n_threads,
    filename,
    output_directory,
    sql,
    side_thread,
    verbose,
    quiet,
):
    """
    Log the reddit submission which ends in LEAF_COMMENT_ID.
    If no comment id is provided, use the latest completed thread found in the thread directory.
    By default, assumes that this is part of the main chain, and will attempt to
    find the true get if the gz or the assist are linked instead.
    """
    from rcounting import configure_logging
    from rcounting import thread_directory as td
    from rcounting import thread_navigation as tn
    from rcounting import utils
    from rcounting.io import ThreadLogger, update_counters_table
    from rcounting.reddit_interface import reddit, subreddit

    t_start = datetime.now()
    utils.ensure_directory(output_directory)

    if side_thread or (
        filename is not None
        and (n_threads != 1 or all_counts or Path(filename).suffix == ".sqlite")
    ):
        sql = True

    configure_logging.setup(printer, verbose, quiet)
    directory = td.load_wiki_page(subreddit, "directory")

    if not leaf_comment_id:
        comment = tn.find_previous_get(directory.rows[0].submission_id)
    else:
        comment = reddit.comment(leaf_comment_id)
    printer.debug(
        "Logging %s reddit submission%s starting at comment id %s and moving backwards",
        "all" if all_counts else n_threads,
        "s" if (n_threads > 1) or all_counts else "",
        comment.id,
    )

    threadlogger = ThreadLogger(sql, output_directory, filename, not side_thread)
    completed = 0

    submission = comment.submission
    submission_id = None
    comment_id = comment.id
    multiple = 1
    while (not all_counts and (completed < n_threads)) or (
        all_counts and submission.id != threadlogger.last_checkpoint
    ):
        printer.info("Logging %s", submission.title)
        if not threadlogger.is_already_logged(submission):
            try:
                if submission_id is not None:
                    comment = tn.find_get_in_submission(submission_id, comment_id)
                df = pd.DataFrame(tn.fetch_comments(comment))
                threadlogger.log(comment, df)
            except TooManyRequests:
                time.sleep(30 * multiple)
                multiple *= 1.5
                continue
        else:
            printer.info("Submission %s has already been logged!", submission.title)

        if submission.id in directory.first_submissions:
            break

        submission_id, comment_id = tn.find_previous_submission(submission)
        submission = reddit.submission(submission_id)
        multiple = 1
        completed += 1

    if completed:
        if sql:
            update_counters_table(threadlogger.db)
        if submission.id in directory.first_submissions + [threadlogger.last_checkpoint]:
            threadlogger.update_checkpoint()
    else:
        printer.info("The database is already up to date!")
    printer.info("Running the script took %s", datetime.now() - t_start)


if __name__ == "__main__":
    log()  # pylint: disable=no-value-for-parameter



import itertools
import math
from collections import Counter, defaultdict
from typing import Sequence, Tuple

import numpy as np
import scipy.sparse

from rcounting import parsing

from .rules import default_rule
from .side_threads import SideThread
from .validate_form import alphanumeric, base_n


class DFA:
    """Generate and store transition matrices for discrete finite automate
    which track what happens when a word from an alphabet of size n_symbols is
    extended by one symbol. Calculating these can be computationally expensive,
    so the code caches them for later use.

    """

    def __init__(self, n_symbols: int, n_states: int, sparse=True):
        self.n_states = n_states
        self.n_symbols = n_symbols
        self.size = self.n_states**self.n_symbols
        self.sparse = sparse
        self.lookup = {str(i): str(min(i + 1, n_states - 1)) for i in range(n_states)}
        self.transitions = None
        self.transition_matrix = None

    def __getitem__(self, i):
        if self.transitions is None:
            self.transitions = [self.generate_identity()]
            self.transition_matrix = self.generate_transition_matrix()
        while len(self.transitions) <= i:
            self.transitions.append(self.transitions[-1] @ self.transition_matrix)

        return self.transitions[i]

    def generate_identity(self):
        if self.sparse:
            return scipy.sparse.eye(self.size, dtype=int, format="csc")
        return np.eye(self.size, dtype=int)

    def _connections(self, i):
        state = np.base_repr(i, self.n_states).zfill(self.n_symbols)
        js = [
            int(state[:ix] + self.lookup[x] + state[ix + 1 :], self.n_states)
            for ix, x in enumerate(state)
        ]
        result = defaultdict(int)
        for j in js:
            if j == 1:
                continue
            result[j] += 1
        return (
            len(result),
            np.array(list(result.keys()), dtype=int),
            np.array(list(result.values()), dtype=int),
        )

    def generate_transition_matrix(self):
        data = np.zeros(self.n_symbols * self.size, dtype=int)
        x = np.zeros(self.n_symbols * self.size, dtype=int)
        y = np.zeros(self.n_symbols * self.size, dtype=int)
        ix = 0
        for i in range(self.size):
            length, js, new_data = self._connections(i)
            x[ix : ix + length] = i
            y[ix : ix + length] = js
            data[ix : ix + length] = new_data
            ix += length
        return scipy.sparse.coo_matrix(
            (data[:ix], (x[:ix], y[:ix])),
            shape=(self.size, self.size),
        ).tocsc()

    def encode(self, state: str) -> int:
        """Converts a word to an integer encoding of the corresponding state vector"""
        counts = Counter(state)
        return sum(
            (self.n_states**pos) * min(self.n_states - 1, counts[digit])
            for pos, digit in enumerate(alphanumeric[: self.n_symbols])
        )


class CompressedDFA(DFA):
    """A more compressed DFA class, where each string is represented by the
    3-tuple of (how many digits don't occur in the string, how many digits
    occur once, how many occur two or more times). This representation
    decreases the number of states to keep track of from ~60k to ~60 for
    10-symbol words"""

    def __init__(self, n_symbols):
        super().__init__(n_symbols=n_symbols, n_states=3)
        self.size = math.comb(n_symbols + 2, n_symbols)
        self.total_lengths = [
            len(range(max(i - n_symbols, 0), i // 2 + 1)) for i in range(2 * n_symbols + 1)
        ]

    def _find_next_states(self, state: Sequence[int]):
        result = []
        if state[0]:
            result.append((state[0], [state[0] - 1, state[1] + 1, state[2]]))
        if state[1]:
            result.append((state[1], [state[0], state[1] - 1, state[2] + 1]))
        if state[2]:
            result.append((state[2], state))
        return result

    def generate_transition_matrix(self):
        transition_matrix = np.zeros((self.size, self.size), dtype=int)
        states = [
            (x, y, self.n_symbols - x - y)
            for x in range(self.n_symbols + 1)
            for y in range(self.n_symbols + 1 - x)
        ]
        for state in states:
            for count, new_state in self._find_next_states(state):
                transition_matrix[self.encode(state), self.encode(new_state)] = count
        return transition_matrix

    def encode(self, state: str | Sequence[int]) -> int:
        if isinstance(state, str):
            counts = Counter([min(x, 2) for x in Counter(state).values()])
        else:
            counts = state
        digit_sum = counts[1] + 2 * counts[2]
        predecessors = sum(self.total_lengths[:digit_sum])
        return predecessors + counts[1] // 2


class LastDigitDFA(DFA):
    """A DFA that keeps track of the last digit of a string, and moves to a
    failure state if the next digit is the same as the current state. Used to
    model the no successive digits side thread

    """

    def __init__(self, n_symbols=10):
        super().__init__(n_symbols=n_symbols, n_states=3, sparse=False)
        self.size = 3

    def generate_transition_matrix(self):
        return np.array([[0, self.n_symbols, 0], [0, self.n_symbols - 1, 0], [0, 0, 0]])

    def encode(self, state: str):
        if not state:
            return 0
        previous = ""
        for char in state:
            if char == previous:
                return 2
            previous = char
        return 1


class NotAnyOfThoseDFA(DFA):
    """A class to represent the states and transitions of the not any of those
    side thread, which has the rule that the cannot match any of:

    - No Repeating Digits
    - Only Repeating Digits
    - Mostly Repeating Digits
    - No Successive Digits
    - No Consecutive Digits
    - Only Consecutive Digits

    We'll use the following representation for a state:

    [a_0, a_1, a_2, \\ldots, a_n], b, c

    where a is a boolean vector, with a_i representing whether the digit i is
    present in the state, b is a count of how many digits are already present twice
    in the state, and c is a ternary digit representing whether:

    0. The the last digit is only present once in the string
    1. The last digit of the string is present at least twice in the string
    2. The state has already successfully failed the NSD condition

    Instead of packing the states nicely in a sensible scheme, we'll just loop
    over all possible values of a, b, c in some arbitrary order, discard the
    values which correspond to invalid states, and use that order to assign an
    integer to each state.

    The invalid states are:

    b = 0, c != 0
    b = sum(a), c == 0

    and after accounting for those, it can be shown that there are 3n*2^(n-1)
    valid states

    """

    def __init__(self, n=10):
        super().__init__(n_symbols=n, n_states=3)
        self.itos_map = self.initialize_itos_map()
        self.stoi_map = {v: k for k, v in self.itos_map.items()}
        self.size = len(self.itos_map)

    def initialize_itos_map(self):
        itos_map = {}
        idx = 0
        for mask in itertools.product([False, True], repeat=self.n_symbols):
            for b in range(sum(mask) + 1):
                for c in range(3):
                    if b == 0 and c > 0:
                        continue
                    if b == sum(mask) and c == 0:
                        continue
                    itos_map[idx] = mask, b, c
                    idx += 1
        return itos_map

    def _find_next_states(self, mask: Tuple[bool, ...], b: int, c: int):
        total = sum(mask)
        result = []
        # Every digit not present in the state just results in one more bit
        # set, the same number of twos in the string, and a last digit of zero
        # if we're not already in a success state for nsd', and two if we are
        for idx, bit in enumerate(mask):
            if not bit:
                current = tuple(mask[:idx]) + (True,) + tuple(mask[idx + 1 :])
                result += [(1, (current, b, 0 if c != 2 else 2))]
        if total > 0:
            # If we aren't already in a success state for \overline{NSD}, then
            # one of the digits we add will move us there
            if c != 2:
                result += [(1, (mask, b + (c == 0), 2))]
            # The remaining digits will just keep the same mask, make b
            # either stay the same or increase by one, and make c = 1 if we are
            # not in a success state for nsd'
            result += [(b - (c == 1), (mask, b, 1 if c != 2 else 2))]
            result += [(total - b - (c == 0), (mask, b + 1, 1 if c != 2 else 2))]
        return [x for x in result if x[0] > 0]

    def int_to_mask(self, state: int) -> Tuple[bool, ...]:
        return tuple(char == "1" for char in f"{state:0>10b}"[::-1])

    def generate_transition_matrix(self):
        i = np.empty(self.n_symbols * self.size, dtype=int)
        j = np.empty(self.n_symbols * self.size, dtype=int)
        data = np.empty(self.n_symbols * self.size, dtype=int)
        idx = 0
        for state in range(self.size):
            next_states = self._find_next_states(*self.itos_map[state])
            counts, new_states = zip(*next_states)
            new_states = [self.stoi_map[ns] for ns in new_states]
            total = len(counts)
            i[idx : idx + total] = state
            j[idx : idx + total] = new_states
            data[idx : idx + total] = counts
            idx += total
        return scipy.sparse.coo_array(
            (data[:idx], (i[:idx], j[:idx])),
            shape=(self.size, self.size),
        ).tocsc()

    def encode(self, state: str) -> int:
        symbols = alphanumeric[: self.n_symbols]
        bitmask = [symbol in state for symbol in symbols]
        current = ""
        c = 0
        for char in state:
            if char == current:
                c = 2
                break
            current = char
        counts = Counter(state)
        if c != 2 and state:
            c = min(counts[state[-1]], 2) - 1
        b = Counter([min(x, 2) for x in counts.values()])[2]
        return self.stoi_map[tuple(bitmask), b, c]


class BarelyRepeatingDigitsDFA(DFA):
    """Class for tracking state transitions for barely repeating digits. We
    have three types of state:

    - One failure state
    - A state representing the empty string
    - States of the form (digits present, repeat), where digits present counts
      how many digits from the alphabet of n symbols are present in the string,
      and repeat is a bool representing whether any of these digits repeats in
      the string.

    """

    def __init__(self, n_symbols=10):
        super().__init__(n_symbols=n_symbols, n_states=3, sparse=False)
        self.size = 2 * (n_symbols + 1)

    def encode(self, state: str | Sequence[int]):
        if isinstance(state, str):
            counts = Counter([min(x, 3) for x in Counter(state).values()])
            digits_present = counts[1] + counts[2]
            repeat = counts[2] if not counts[3] else -1
        else:
            digits_present, repeat = state

        if digits_present > self.n_symbols or digits_present < 0 or repeat not in [0, 1]:
            return 0
        if digits_present == 0:
            return 1 - repeat
        return 2 * digits_present + repeat

    def _find_next_states(self, state: Tuple[int, int]):
        digits_present, repeat = state
        result = []
        if repeat == 0:
            result.append((digits_present, (digits_present, 1)))
        if digits_present != self.n_symbols:
            result.append((self.n_symbols - digits_present, (digits_present + 1, repeat)))
        return result

    def generate_transition_matrix(self):
        transition_matrix = np.zeros((self.size, self.size), dtype=int)
        states = [
            (digits_present, repeat)
            for digits_present in range(self.n_symbols + 1)
            for repeat in range(2)
        ]
        for state in states:
            for count, new_state in self._find_next_states(state):
                transition_matrix[self.encode(state), self.encode(new_state)] = count
        transition_matrix[0] = 0
        transition_matrix[:, 0] = 0
        return transition_matrix


class DFAThread(SideThread):
    """Describing side threads using a deterministic finite automaton.

    A lot of side threads have rules like "valid counts are those were every
    digit repeats at least once" or "valid counts are made up of a set of
    consecutive digits". Determining how many valid counts there are smaller
    than or equal to a given number is tricky to do, but that's exactly what we
    need in order to convert a comment to a count.

    These threads have the property that the validity of a count only depends
    on which digits are present in a comment, and not on the order in which
    they appear. That means that we can describe the state vector of a given
    comment by the tuple of digit counts.

    We can then describe what happens to the state when a given digit is
    appended to the count -- the corresponding entry in the state tuple is
    increased by one, or, if the entry is already at some maximal value, the
    entry is just kept constant. For example, for only repeating digits the
    three states we are interested in are:

    - digit occurs 0 times
    - digit occurs once
    - digit occurs 2 or more times

    and after appending a digit, the possible new states for that digit are

    - digit occurs once (if it was not present before)
    - digit occurs 2 or more times (if it was present before)

    Once we have a description of the possible new states for any given state
    after appending an arbitrary digit, we are basically done: we can start
    with a given input states, apply the transition a certain number of times,
    and see how many of the states we end up with follow whatever rules we've set up.

    See
    https://cstheory.stackexchange.com/questions/8200/counting-words-accepted-by-a-regular-grammar/8202#8202
    for more of a description of how it works.

    The rule and form attributes of the side threads are the same as for base
    n; no validation that each digit actually occurs the correct number of
    times is currently done.

    """

    def __init__(
        self,
        dfa_base=3,
        n=10,
        rule=default_rule,
        dfa: DFA | None = None,
        indices: Sequence[int] | None = None,
        offset=0,
    ):
        self.n = n
        form = base_n(n)
        self.dfa = dfa if dfa is not None else DFA(n, dfa_base)
        self.indices = indices if indices is not None else []
        self.encoded_symbols = [self.dfa.encode(s) for s in alphanumeric[:n]]
        self.matrices = {}

        # Some of the threads skip the single-digit counts which would
        # otherwhise be valid, so we add an offset to account for that
        self.offset = offset

        super().__init__(rule=rule, form=form, comment_to_count=self.count)

    def matrix(self, n):
        if n not in self.matrices:
            matrix = self.dfa[n][:, self.indices]
            if scipy.sparse.issparse(matrix):
                matrix = matrix.toarray()
            self.matrices[n] = matrix.sum(axis=1)
        return self.matrices[n]

    def word_is_valid(self, word):
        return self.dfa.encode(word) in self.indices

    def count(self, comment_body: str, bijective=False) -> int:
        word = parsing.extract_count_string(comment_body, self.n).lower()
        word_length = len(word)

        enumeration = 0
        for i in range(word_length - 1, -1, -1):
            matrix_power = word_length - 1 - i
            prefix = word[:i]
            current_char = word[i]
            suffixes = alphanumeric[i == 0 : alphanumeric.index(current_char)]
            states = [self.dfa.encode(prefix + suffix) for suffix in suffixes]
            if bijective:
                states += [self.dfa.encode("")]
            elif matrix_power > 0:
                enumeration += self._enumerate(self.encoded_symbols[1:], matrix_power - 1)
            enumeration += self._enumerate(states, matrix_power)
        return enumeration + self.word_is_valid(word) - self.offset

    def _enumerate(self, states, n):
        """Given a list of states, how many success strings will we have after
        n rounds of appending digits?"""
        if not states:
            return 0
        states, counts = zip(*Counter(states).items())
        return np.dot(counts, self.matrix(n)[list(states)])


dfa_10_2 = DFA(10, 2)
compressed_dfa = CompressedDFA(10)


def no_consecutive_states(n_symbols):
    """The valid states in no consecutive digits threads, encoded as binary
    strings where the i'th digit of the string indicates whether the
    corresponding digit is present in the original string or not.

    For example, 0000000000 is the empty string, while 1000000000 represents
    strings which only contain the digit 9

    """
    current = ["0", "01"]
    result = []
    while current:
        val = current.pop()
        if len(val) < n_symbols:
            current.append("0" + val)
            current.append("01" + val)
        elif len(val) == n_symbols:
            result.append(val)
        else:
            result.append(val[1:])
    return result


no_consecutive = sorted([int(x, 2) for x in no_consecutive_states(10)])[1:]
no_successive = [1]
no_repeating = [compressed_dfa.encode((x, 10 - x, 0)) for x in range(10)]
only_repeating = [compressed_dfa.encode((x, 0, 10 - x)) for x in range(10)]
mostly_repeating = [compressed_dfa.encode((x, 1, 10 - x - 1)) for x in range(10 - 1)]

# The only consecutive indices are sums of adjacent powers of two
only_consecutive = sorted(
    [
        sum(2**k for k in range(minval, maxval))
        for maxval in range(10 + 1)
        for minval in range(maxval)
    ]
)

not_any_dfa = NotAnyOfThoseDFA()
not_any_mask = [
    not_any_dfa.int_to_mask(x)
    for x in range(1024)
    if x not in only_consecutive and x not in no_consecutive
]
# Valid states for not any have a bitmask that excludes dfa
not_any = sorted(
    [not_any_dfa.stoi_map[tuple(a), b, 2] for a in not_any_mask for b in range(1, sum(a) - 1)]
)

barely_repeating_dfa = BarelyRepeatingDigitsDFA()
barely_repeating = [barely_repeating_dfa.encode((x, 1)) for x in range(2, 11)]

dfa_threads = {
    "mostly repeating digits": DFAThread(dfa=compressed_dfa, indices=mostly_repeating),
    "no consecutive digits": DFAThread(dfa=dfa_10_2, indices=no_consecutive),
    "no repeating digits": DFAThread(dfa=compressed_dfa, indices=no_repeating),
    "no successive digits": DFAThread(indices=no_successive, dfa=LastDigitDFA()),
    "only consecutive digits": DFAThread(dfa=dfa_10_2, indices=only_consecutive, offset=9),
    "only repeating digits": DFAThread(dfa=compressed_dfa, indices=only_repeating),
    "not any of those": DFAThread(dfa=not_any_dfa, indices=not_any),
    "barely repeating digits": DFAThread(dfa=barely_repeating_dfa, indices=barely_repeating),
}
